import numpy as np
from numpy.typing import NDArray

from .analysis_synthesis import AnalysisSynthesisTSM
from .array import ArrReader, ArrWriter


def hanning(length: int) -> np.ndarray:
    time = np.arange(length)
    return 0.5 * (1 - np.cos(2 * np.pi * time / length))


def phasevocoder(
    channels: int, speed: float, arr: np.ndarray, frame_length: int = 2048
) -> NDArray[np.int16]:

    # Frame length should be a power of two for maximum performance.

    synthesis_hop = frame_length // 4
    analysis_hop = int(synthesis_hop * speed)

    analysis_window = hanning(frame_length)
    synthesis_window = hanning(frame_length)

    writer = ArrWriter(np.zeros((0, channels), dtype=np.int16))
    reader = ArrReader(arr)

    AnalysisSynthesisTSM(
        channels,
        frame_length,
        analysis_hop,
        synthesis_hop,
        analysis_window,
        synthesis_window,
    ).run(reader, writer)

    return writer.output



from dataclasses import dataclass

from auto_editor.utils.types import AlignType

# start - When the clip starts in the timeline
# dur - The duration of the clip in the timeline before speed is applied
# offset - When from the source to start playing the media at


@dataclass
class VideoObj:
    start: int
    dur: int
    offset: int
    speed: float
    src: int
    stream: int = 0


@dataclass
class AudioObj:
    start: int
    dur: int
    offset: int
    speed: float
    src: int
    stream: int = 0


@dataclass
class TextObj:
    start: int
    dur: int
    content: str
    x: int = "centerX"  # type: ignore
    y: int = "centerY"  # type: ignore
    size: int = 55
    font: str = "Arial"
    align: AlignType = "left"
    fill: str = "#FFF"
    stroke: int = 0
    strokecolor: str = "#000"


@dataclass
class ImageObj:
    start: int
    dur: int
    src: str
    x: int = "centerX"  # type: ignore
    y: int = "centerY"  # type: ignore
    opacity: float = 1
    anchor: str = "ce"
    rotate: float = 0  # in degrees


@dataclass
class RectangleObj:
    start: int
    dur: int
    x: int
    y: int
    width: int
    height: int
    anchor: str = "ce"
    fill: str = "#c4c4c4"
    stroke: int = 0
    strokecolor: str = "#000"


@dataclass
class EllipseObj:
    start: int
    dur: int
    x: int
    y: int
    width: int
    height: int
    anchor: str = "ce"
    fill: str = "#c4c4c4"
    stroke: int = 0
    strokecolor: str = "#000"



from auto_editor.timeline import Timeline
from auto_editor.utils.func import aspect_ratio, to_timecode


def timecode_to_frames(timecode: str, fps: float) -> int:
    _h, _m, _s = timecode.split(":")
    h = int(_h)
    m = int(_m)
    s = float(_s)
    return round((h * 3600 + m * 60 + s) * fps)


def shotcut_xml(
    output: str,
    timeline: Timeline,
) -> None:
    width, height = timeline.res
    num, den = aspect_ratio(width, height)

    chunks = timeline.chunks
    if chunks is None:
        raise ValueError("Timeline too complex")
    fps = timeline.fps
    inp = timeline.inp

    global_out = to_timecode(timeline.out_len() / fps, "standard")
    version = "21.05.18"

    with open(output, "w", encoding="utf-8") as out:
        out.write('<?xml version="1.0" standalone="no"?>\n')
        out.write(
            '<mlt LC_NUMERIC="C" version="7.1.0" '
            f'title="Shotcut version {version}" producer="main_bin">\n'
        )
        out.write(
            '\t<profile description="automatic" '
            f'width="{width}" height="{height}" '
            'progressive="1" sample_aspect_num="1" sample_aspect_den="1" '
            f'display_aspect_num="{num}" display_aspect_den="{den}" '
            f'frame_rate_num="{fps}" frame_rate_den="1" colorspace="709"/>\n'
        )
        out.write('\t<playlist id="main_bin">\n')
        out.write('\t\t<property name="xml_retain">1</property>\n')
        out.write("\t</playlist>\n")

        # out was the new video length in the original xml
        out.write(f'\t<producer id="black" in="00:00:00.000" out="{global_out}">\n')
        out.write(f'\t\t<property name="length">{global_out}</property>\n')
        out.write('\t\t<property name="eof">pause</property>\n')
        out.write('\t\t<property name="resource">0</property>\n')
        out.write('\t\t<property name="aspect_ratio">1</property>\n')
        out.write('\t\t<property name="mlt_service">color</property>\n')
        out.write('\t\t<property name="mlt_image_format">rgba</property>\n')
        out.write('\t\t<property name="set.test_audio">0</property>\n')
        out.write("\t</producer>\n")

        out.write('\t<playlist id="background">\n')  # same for this out too.
        out.write(
            f'\t\t<entry producer="black" in="00:00:00.000" out="{global_out}"/>\n'
        )
        out.write("\t</playlist>\n")

        chains = 0
        producers = 0

        # Speeds like [1.5, 3] don't work because of duration issues, too bad!

        for clip in chunks:
            if clip[2] == 99999:
                continue

            speed = clip[2]

            _out = to_timecode(clip[1] / speed / fps, "standard")
            length = to_timecode((clip[1] / speed + 1) / fps, "standard")

            if speed == 1:
                resource = inp.path
                caption = inp.basename
                out.write(f'\t<chain id="chain{chains}" out="{_out}">\n')
            else:
                resource = f"{speed}:{inp.path}"
                caption = f"{inp.basename} ({speed}x)"
                out.write(
                    f'\t<producer id="producer{producers}" in="00:00:00.000" out="{_out}">\n'
                )
                producers += 1
            chains += 1

            out.write(f'\t\t<property name="length">{length}</property>\n')
            out.write('\t\t<property name="eof">pause</property>\n')
            out.write(f'\t\t<property name="resource">{resource}</property>\n')

            if speed == 1:
                out.write(
                    '\t\t<property name="mlt_service">avformat-novalidate</property>\n'
                )
                out.write('\t\t<property name="seekable">1</property>\n')
                out.write('\t\t<property name="audio_index">1</property>\n')
                out.write('\t\t<property name="video_index">0</property>\n')
                out.write('\t\t<property name="mute_on_pause">0</property>\n')
                out.write(
                    f'\t\t<property name="shotcut:caption">{caption}</property>\n'
                )
                out.write('\t\t<property name="xml">was here</property>\n')
            else:
                out.write('\t\t<property name="aspect_ratio">1</property>\n')
                out.write('\t\t<property name="seekable">1</property>\n')
                out.write('\t\t<property name="audio_index">1</property>\n')
                out.write('\t\t<property name="video_index">0</property>\n')
                out.write('\t\t<property name="mute_on_pause">1</property>\n')
                out.write(f'\t\t<property name="warp_speed">{speed}</property>\n')
                out.write(f'\t\t<property name="warp_resource">{inp.path}</property>\n')
                out.write('\t\t<property name="mlt_service">timewarp</property>\n')
                out.write('\t\t<property name="shotcut:producer">avformat</property>\n')
                out.write('\t\t<property name="video_delay">0</property>\n')
                out.write(
                    f'\t\t<property name="shotcut:caption">{caption}</property>\n'
                )
                out.write('\t\t<property name="xml">was here</property>\n')
                out.write('\t\t<property name="warp_pitch">1</property>\n')

            out.write("\t</chain>\n" if speed == 1 else "\t</producer>\n")

        out.write('\t<playlist id="playlist0">\n')
        out.write('\t\t<property name="shotcut:video">1</property>\n')
        out.write('\t\t<property name="shotcut:name">V1</property>\n')

        producers = 0
        i = 0
        for clip in chunks:
            if clip[2] == 99999:
                continue

            speed = clip[2]

            if speed == 1:
                in_len: float = clip[0] - 1
            else:
                in_len = max(clip[0] / speed, 0)

            out_len = max((clip[1] - 2) / speed, 0)

            _in = to_timecode(in_len / fps, "standard")
            _out = to_timecode(out_len / fps, "standard")

            tag_name = f"chain{i}"
            if speed != 1:
                tag_name = f"producer{producers}"
                producers += 1

            out.write(f'\t\t<entry producer="{tag_name}" in="{_in}" out="{_out}"/>\n')
            i += 1

        out.write("\t</playlist>\n")

        out.write(
            f'\t<tractor id="tractor0" title="Shotcut version {version}" '
            f'in="00:00:00.000" out="{global_out}">\n'
        )
        out.write('\t\t<property name="shotcut">1</property>\n')
        out.write('\t\t<property name="shotcut:projectAudioChannels">2</property>\n')
        out.write('\t\t<property name="shotcut:projectFolder">0</property>\n')
        out.write('\t\t<track producer="background"/>\n')
        out.write('\t\t<track producer="playlist0"/>\n')
        out.write('\t\t<transition id="transition0">\n')
        out.write('\t\t\t<property name="a_track">0</property>\n')
        out.write('\t\t\t<property name="b_track">1</property>\n')
        out.write('\t\t\t<property name="mlt_service">mix</property>\n')
        out.write('\t\t\t<property name="always_active">1</property>\n')
        out.write('\t\t\t<property name="sum">1</property>\n')
        out.write("\t\t</transition>\n")
        out.write('\t\t<transition id="transition1">\n')
        out.write('\t\t\t<property name="a_track">0</property>\n')
        out.write('\t\t\t<property name="b_track">1</property>\n')
        out.write('\t\t\t<property name="version">0.9</property>\n')
        out.write('\t\t\t<property name="mlt_service">frei0r.cairoblend</property>\n')
        out.write('\t\t\t<property name="threads">0</property>\n')
        out.write('\t\t\t<property name="disable">1</property>\n')
        out.write("\t\t</transition>\n")

        out.write("\t</tractor>\n")
        out.write("</mlt>\n")



import numpy as np
from numpy.typing import NDArray


class ArrReader:
    __slots__ = ("samples", "pointer")

    def __init__(self, arr: np.ndarray) -> None:
        self.samples = arr
        self.pointer = 0

    @property
    def empty(self) -> bool:
        return self.samples.shape[0] <= self.pointer

    def read(self, buffer: np.ndarray) -> int:
        end = self.pointer + buffer.shape[1]
        frames = self.samples[self.pointer : end].T.astype(np.float32)
        n = frames.shape[1]
        np.copyto(buffer[:, :n], frames)
        del frames
        self.pointer = end
        return n

    def skip(self, n: int) -> int:
        self.pointer += n
        return n


class ArrWriter:
    __slots__ = ("output", "pointer")

    def __init__(self, arr: NDArray[np.int16]) -> None:
        self.output = arr
        self.pointer = 0

    def write(self, buffer: np.ndarray) -> int:
        end = self.pointer + buffer.shape[1]
        changed_buffer: NDArray[np.int16] = buffer.T.astype(np.int16)
        self.output = np.concatenate((self.output, changed_buffer))
        self.pointer = end

        return buffer.shape[1]



import sys
from math import floor
from platform import system
from shutil import get_terminal_size
from time import localtime, time
from typing import Union

from .func import get_stdout


class ProgressBar:
    def __init__(self, bar_type: str) -> None:

        self.machine = False
        self.hide = False

        self.icon = "⏳"
        self.chars = [" ", "▏", "▎", "▍", "▌", "▋", "▊", "▉", "█"]
        self.brackets = ("|", "|")

        if bar_type == "classic":
            self.icon = "⏳"
            self.chars = ["░", "█"]
            self.brackets = ("[", "]")
        if bar_type == "ascii":
            self.icon = "& "
            self.chars = ["-", "#"]
            self.brackets = ("[", "]")
        if bar_type == "machine":
            self.machine = True
        if bar_type == "none":
            self.hide = True

        self.part_width = len(self.chars) - 1

        self.ampm = True
        if system() == "Darwin" and bar_type in ("default", "classic"):
            try:
                date_format = get_stdout(
                    ["defaults", "read", "com.apple.menuextra.clock", "DateFormat"]
                )
                self.ampm = "a" in date_format
            except FileNotFoundError:
                pass

    @staticmethod
    def pretty_time(my_time: float, ampm: bool) -> str:
        new_time = localtime(my_time)

        hours = new_time.tm_hour
        minutes = new_time.tm_min

        if ampm:
            if hours == 0:
                hours = 12
            if hours > 12:
                hours -= 12
            ampm_marker = "PM" if new_time.tm_hour >= 12 else "AM"
            return f"{hours:02}:{minutes:02} {ampm_marker}"
        return f"{hours:02}:{minutes:02}"

    def tick(self, index: Union[int, float]) -> None:

        if self.hide:
            return

        progress = min(1, max(0, index / self.total))

        if progress == 0:
            progress_rate = 0.0
        else:
            progress_rate = (time() - self.begin_time) / progress

        if self.machine:
            index = min(index, self.total)
            raw = int(self.begin_time + progress_rate)
            print(
                f"{self.title}~{index}~{self.total}~{self.begin_time}~{raw}",
                end="\r",
                flush=True,
            )
            return

        new_time = self.pretty_time(self.begin_time + progress_rate, self.ampm)

        percent = round(progress * 100, 1)
        p_pad = " " * (4 - len(str(percent)))

        columns = get_terminal_size().columns
        bar_len = max(1, columns - (self.len_title + 32))

        progress_bar_str = self.progress_bar_str(progress, bar_len)

        bar = f"  {self.icon}{self.title} {progress_bar_str} {p_pad}{percent}%  ETA {new_time}"

        if len(bar) > columns - 2:
            bar = bar[: columns - 2]
        else:
            bar += " " * (columns - len(bar) - 4)

        sys.stdout.write(bar + "\r")

    def start(self, total: Union[int, float], title: str = "Please wait") -> None:
        self.title = title
        self.len_title = len(title)
        self.total = total
        self.begin_time = time()

        try:
            self.tick(0)
        except UnicodeEncodeError:
            self.icon = "& "
            self.chars = ["-", "#"]
            self.brackets = ("[", "]")
            self.part_width = 1

    def progress_bar_str(self, progress: float, width: int) -> str:
        whole_width = floor(progress * width)
        remainder_width = (progress * width) % 1
        part_width = floor(remainder_width * self.part_width)
        part_char = self.chars[part_width]

        if width - whole_width - 1 < 0:
            part_char = ""

        line = (
            self.brackets[0]
            + self.chars[-1] * whole_width
            + part_char
            + self.chars[0] * (width - whole_width - 1)
            + self.brackets[1]
        )
        return line

    @staticmethod
    def end() -> None:
        sys.stdout.write(" " * (get_terminal_size().columns - 2) + "\r")



"""
Make a pre-edited file reference that can be inputted back into auto-editor.
"""

import json
import os
import sys
from typing import Any, Optional, Union

from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.timeline import Timeline, clipify, make_av
from auto_editor.utils.log import Log
from auto_editor.utils.types import Chunks


def check_attrs(data: object, log: Log, *attrs: str) -> None:
    if not isinstance(data, dict):
        log.error("Data is in wrong shape!")
    for attr in attrs:
        if attr not in data:
            log.error(f"'{attr}' attribute not found!")


def check_file(path: str, log: Log):
    if not os.path.isfile(path):
        log.error(f"Could not locate media file: '{path}'")


def validate_chunks(chunks: object, log: Log) -> Chunks:
    if not isinstance(chunks, (list, tuple)):
        log.error("Chunks must be a list")

    if len(chunks) == 0:
        log.error("Chunks are empty!")

    new_chunks = []
    prev_end: Optional[int] = None

    for i, chunk in enumerate(chunks):
        if len(chunk) != 3:
            log.error("Chunk must have a length of 3.")

        if i == 0 and chunk[0] != 0:
            log.error("First chunk must start with 0")

        if chunk[1] - chunk[0] < 1:
            log.error("Chunk duration must be at least 1")

        if chunk[2] <= 0 or chunk[2] > 99999:
            log.error("Chunk speed range must be >0 and <=99999")

        if prev_end is not None and chunk[0] != prev_end:
            log.error(f"Chunk disjointed at {chunk}")

        prev_end = chunk[1]

        new_chunks.append((chunk[0], chunk[1], float(chunk[2])))

    return new_chunks


class Version:
    __slots__ = ("major", "minor", "micro")

    def __init__(self, val: str, log: Log) -> None:
        ver_str = val.split(".")
        if len(ver_str) > 3:
            log.error("Version string: Too many separators!")
        while len(ver_str) < 3:
            ver_str.append("0")

        try:
            self.major, self.minor, self.micro = map(int, ver_str)
        except ValueError:
            log.error("Version string: Could not convert to int.")

    def __eq__(self, other: object) -> bool:
        if isinstance(other, tuple) and len(other) == 2:
            return (self.major, self.minor) == other
        return (self.major, self.minor, self.micro) == other

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.micro}"


def read_json(path: str, ffmpeg: FFmpeg, log: Log) -> Timeline:
    with open(path, "r") as f:
        data = json.load(f)

    check_attrs(data, log, "version")
    version = Version(data["version"], log)

    if version == (1, 0) or version == (0, 1):
        check_attrs(data, log, "source", "chunks")
        check_file(data["source"], log)

        chunks = validate_chunks(data["chunks"], log)
        inp = FileInfo(data["source"], ffmpeg, log)

        vspace, aspace = make_av(clipify(chunks, 0), inp)

        fps = inp.get_fps()
        sr = inp.get_samplerate()
        res = inp.get_res()

        return Timeline([inp], fps, sr, res, "#000", vspace, aspace, chunks)

    if version == (2, 0) or version == (0, 2):
        check_attrs(data, log, "timeline")
        # check_file(data["source"], log)
        # return data["background"], data["source"], chunks

        raise ValueError("Incomplete")

    log.error(f"Unsupported version: {version}")


def make_json_timeline(
    _version: str,
    out: Union[str, int],
    timeline: Timeline,
    log: Log,
) -> None:

    version = Version(_version, log)

    if version == (1, 0) or version == (0, 1):
        if timeline.chunks is None:
            log.error("Timeline too complex to convert to version 1.0")

        data: Any = {
            "version": "1.0.0",
            "source": os.path.abspath(timeline.inp.path),
            "chunks": timeline.chunks,
        }
    elif version == (2, 0) or version == (0, 2):
        sources = [os.path.abspath(inp.path) for inp in timeline.inputs]
        data = {
            "version": "2.0.0",
            "sources": sources,
            "timeline": {
                "background": timeline.background,
                "resolution": timeline.res,
                "fps": timeline.fps,
                "samplerate": timeline.samplerate,
                "video": timeline.v,
                "audio": timeline.a,
            },
        }
    else:
        log.error(f"Version {version} is not supported!")

    if isinstance(out, str):
        if not out.endswith(".json"):
            log.error("Output extension must be .json")

        with open(out, "w") as outfile:
            json.dump(data, outfile, indent=2, default=lambda o: o.__dict__)
    else:
        json.dump(data, sys.stdout, indent=2, default=lambda o: o.__dict__)
        print("")  # Flush stdout



import os
import re
import subprocess
from platform import system
from typing import List

from auto_editor.ffwrapper import FFmpeg
from auto_editor.utils.func import get_stdout
from auto_editor.utils.log import Log


def get_domain(url: str) -> str:
    from urllib.parse import urlparse

    t = urlparse(url).netloc
    return ".".join(t.split(".")[-2:])


def download_video(my_input: str, args, ffmpeg: FFmpeg, log: Log) -> str:
    log.conwrite("Downloading video...")

    download_format = args.download_format

    if download_format is None and get_domain(my_input) == "youtube.com":
        download_format = "bestvideo[ext=mp4]+bestaudio[ext=m4a]"

    if args.output_format is None:
        output_format = re.sub(r"\W+", "-", os.path.splitext(my_input)[0]) + ".%(ext)s"
    else:
        output_format = args.output_format

    yt_dlp_path = args.yt_dlp_location

    cmd = ["--ffmpeg-location", ffmpeg.path]

    if download_format is not None:
        cmd.extend(["-f", download_format])

    cmd.extend(["-o", output_format, my_input])

    if args.yt_dlp_extras is not None:
        cmd.extend(args.yt_dlp_extras.split(" "))

    try:
        location = get_stdout(
            [yt_dlp_path, "--get-filename", "--no-warnings"] + cmd
        ).strip()
    except FileNotFoundError:
        msg = "Could not find program 'yt-dlp' when attempting to download a URL. Install yt-dlp with "
        if system() == "Windows":
            msg += "your favorite package manager (pip, choco, winget)."
        elif system() == "Darwin":
            msg += "brew or pip and make sure it's in PATH."
        else:
            msg += "pip or your favorite package manager and make sure it's in PATH."
        log.error(msg)

    if not os.path.isfile(location):
        subprocess.run([yt_dlp_path] + cmd)

    if not os.path.isfile(location):
        log.error(f"Download file wasn't created: {location}")

    return location


def valid_input(inputs: List[str], ffmpeg: FFmpeg, args, log: Log) -> List[str]:
    new_inputs = []

    for my_input in inputs:
        if os.path.isfile(my_input):
            _, ext = os.path.splitext(my_input)
            if ext == "":
                log.error("File must have an extension.")
            new_inputs.append(my_input)

        elif my_input.startswith("http://") or my_input.startswith("https://"):
            new_inputs.append(download_video(my_input, args, ffmpeg, log))
        else:
            if os.path.isdir(my_input):
                log.error("Input must be a file or a URL, not a directory.")
            log.error(f"Could not find file: '{my_input}'")

    return new_inputs



def safe_mkdir(path: str) -> str:
    from os import mkdir
    from shutil import rmtree

    try:
        mkdir(path)
    except OSError:
        rmtree(path)
        mkdir(path)
    return path


def indent(base: int, *lines: str) -> str:
    new_lines = ""
    for line in lines:
        new_lines += ("\t" * base) + line + "\n"
    return new_lines



import av
import numpy as np
from numpy.typing import NDArray
from PIL import ImageChops

from auto_editor.utils.progressbar import ProgressBar


def pixel_difference(
    path: str, fps: float, progress: ProgressBar
) -> NDArray[np.uint64]:
    container = av.open(path, "r")

    stream = container.streams.video[0]
    stream.thread_type = "AUTO"

    inaccurate_dur = int(stream.duration * stream.time_base * stream.rate)

    progress.start(inaccurate_dur, "Analyzing pixel diffs")

    prev_image = None
    image = None
    index = 0

    threshold_list = np.zeros((1024), dtype=np.uint64)

    for frame in container.decode(stream):
        if image is None:
            prev_image = None
        else:
            prev_image = image

        index = int(frame.time * fps)
        progress.tick(index)

        if index > len(threshold_list) - 1:
            threshold_list = np.concatenate(
                (threshold_list, np.zeros((len(threshold_list)), dtype=np.uint64)),
                axis=0,
            )

        image = frame.to_image()

        if prev_image is not None:
            threshold_list[index] = np.count_nonzero(
                ImageChops.difference(prev_image, image)
            )

    progress.end()
    return threshold_list[:index]



import json
import os.path
import subprocess
import sys
from dataclasses import dataclass
from fractions import Fraction
from platform import system
from re import search
from subprocess import PIPE, Popen
from typing import List, Optional, Tuple

from auto_editor.utils.func import get_stdout, to_timecode
from auto_editor.utils.log import Log

IMG_CODECS = ("png", "mjpeg", "webp")
SUB_EXTS = {"mov_text": "srt", "ass": "ass", "webvtt": "vtt"}


class FFmpeg:
    __slots__ = ("debug", "path", "version")

    @staticmethod
    def _set_ff_path(ff_location: Optional[str], my_ffmpeg: bool) -> str:
        if ff_location is not None:
            return ff_location
        if my_ffmpeg or system() not in ("Windows", "Darwin"):
            return "ffmpeg"

        # Assumming accessing __file__ is somewhat expensive.
        program = "ffmpeg.exe" if system() == "Windows" else "ffmpeg"
        dirpath = os.path.dirname(os.path.realpath(__file__))
        file_path = os.path.join(dirpath, "ffmpeg", system(), program)

        # May not exist because auto-editor doesn't bundle every architecture.
        if os.path.isfile(file_path):
            return file_path

        return "ffmpeg"

    def __init__(
        self,
        ff_location: Optional[str] = None,
        my_ffmpeg: bool = False,
        debug: bool = False,
    ) -> None:

        self.debug = debug
        self.path = self._set_ff_path(ff_location, my_ffmpeg)
        try:
            _version = get_stdout([self.path, "-version"]).split("\n")[0]
            _version = _version.replace("ffmpeg version", "").strip()
            self.version = _version.split(" ")[0]
        except FileNotFoundError:
            if system() == "Darwin":
                Log().error(
                    "No ffmpeg found, download via homebrew or restore the "
                    "included binary."
                )
            if system() == "Windows":
                Log().error(
                    "No ffmpeg found, download ffmpeg with your favorite package "
                    "manager (ex chocolatey), or restore the included binary."
                )

            Log().error("ffmpeg must be installed and on PATH.")

    def print(self, message: str) -> None:
        if self.debug:
            sys.stderr.write(f"FFmpeg: {message}\n")

    def print_cmd(self, cmd: List[str]) -> None:
        if self.debug:
            sys.stderr.write(f"FFmpeg run: {' '.join(cmd)}\n")

    def run(self, cmd: List[str]) -> None:
        cmd = [self.path, "-y", "-hide_banner"] + cmd
        if not self.debug:
            cmd.extend(["-nostats", "-loglevel", "error"])
        self.print_cmd(cmd)
        subprocess.run(cmd)

    def run_check_errors(
        self,
        cmd: List[str],
        log: Log,
        show_out: bool = False,
        path: Optional[str] = None,
    ) -> None:
        def _run(cmd: List[str]) -> str:
            process = self.Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)

            _, stderr = process.communicate()

            if process.stdin is not None:
                process.stdin.close()
            return stderr.decode("utf-8", "replace")

        output = _run(cmd)

        if "Try -allow_sw 1" in output:
            cmd.insert(-1, "-allow_sw")
            cmd.insert(-1, "1")
            output = _run(cmd)

        error_list = [
            r"Unknown encoder '.*'",
            r"-q:v qscale not available for encoder\. Use -b:v bitrate instead\.",
            r"Specified sample rate .* is not supported",
            r'Unable to parse option value ".*"',
            r"Error setting option .* to value .*\.",
            r"Undefined constant or missing '.*' in '.*'",
            r"DLL .* failed to open",
            r"Incompatible pixel format '.*' for codec '[A-Za-z0-9_]*'",
            r"Unrecognized option '.*'",
            r"Permission denied",
        ]

        if self.debug:
            print(f"stderr: {output}")

        for item in error_list:
            check = search(item, output)
            if check:
                log.error(check.group())

        if path is not None and not os.path.isfile(path):
            log.error(f"The file {path} was not created.")
        elif show_out and not self.debug:
            print(f"stderr: {output}")

    def Popen(self, cmd: List[str], stdin=None, stdout=PIPE, stderr=None) -> Popen:
        cmd = [self.path] + cmd
        self.print_cmd(cmd)
        return Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr)

    def pipe(self, cmd: List[str]) -> str:
        cmd = [self.path, "-y"] + cmd

        self.print_cmd(cmd)
        output = get_stdout(cmd)
        self.print(output)
        return output


@dataclass
class VideoStream:
    width: int
    height: int
    codec: str
    fps: float
    bitrate: Optional[str]
    lang: Optional[str]


@dataclass
class AudioStream:
    codec: str
    samplerate: int
    bitrate: Optional[str]
    lang: Optional[str]


@dataclass
class SubtitleStream:
    codec: str
    ext: str
    lang: Optional[str]


class FileInfo:
    __slots__ = (
        "path",
        "abspath",
        "basename",
        "dirname",
        "name",
        "ext",
        "bitrate",
        "description",
        "videos",
        "audios",
        "subtitles",
    )

    def get_res(self) -> Tuple[int, int]:
        if len(self.videos) > 0:
            return self.videos[0].width, self.videos[0].height
        return 1920, 1080

    def get_fps(self) -> float:
        fps = None
        if len(self.videos) > 0:
            fps = self.videos[0].fps

        return 30 if fps is None else fps

    def get_samplerate(self) -> int:
        if len(self.audios) > 0:
            return self.audios[0].samplerate
        return 48000

    def __init__(self, path: str, ffmpeg: FFmpeg, log: Log):
        self.path = path
        self.abspath = os.path.abspath(path)
        self.basename = os.path.basename(path)
        self.dirname = os.path.dirname(os.path.abspath(path))
        self.name, self.ext = os.path.splitext(path)

        self.videos: List[VideoStream] = []
        self.audios: List[AudioStream] = []
        self.subtitles: List[SubtitleStream] = []
        self.description = None

        _dir = os.path.dirname(ffmpeg.path)
        _ext = os.path.splitext(ffmpeg.path)[1]
        ffprobe = os.path.join(_dir, f"ffprobe{_ext}")

        try:
            info = get_stdout(
                [
                    ffprobe,
                    "-v",
                    "-8",
                    "-print_format",
                    "json",
                    "-show_streams",
                    "-show_format",
                    path,
                ]
            )
        except FileNotFoundError:
            log.error(f"Could not find: {ffprobe}")

        try:
            json_info = json.loads(info)
            if "streams" not in json_info:
                raise ValueError("Key 'streams' not found")
            if "format" not in json_info:
                raise ValueError("Key 'format' not found")
        except Exception as e:
            log.error(f"{path}: Could not read ffprobe JSON: {e}")

        self.bitrate = json_info["format"]["bit_rate"]
        if (
            "tags" in json_info["format"]
            and "description" in json_info["format"]["tags"]
        ):
            self.description = json_info["format"]["tags"]["description"]

        for stream in json_info["streams"]:
            lang = None
            br = None
            if "tags" in stream and "language" in stream["tags"]:
                lang = stream["tags"]["language"]
            if "bit_rate" in stream:
                br = stream["bit_rate"]

            if "codec_type" not in stream:
                log.error("'codec_type' must be in ffprobe json")
            codec_type = stream["codec_type"]
            if not isinstance(codec_type, str):
                log.error("'codec_type' must be a string")

            if codec_type in ("video", "audio", "subtitle"):
                if "codec_name" not in stream:
                    log.error("'codec_name' must be in ffprobe json")
                codec = stream["codec_name"]

                if not isinstance(codec, str):
                    log.error("'codec_name' must be a string")

            if codec_type == "video":
                codec = stream["codec_name"]
                try:
                    fps = float(Fraction(stream["avg_frame_rate"]))
                except ZeroDivisionError:
                    fps = 0
                except ValueError:
                    log.error(
                        f"Could not convert fps '{stream['avg_frame_rate']}' to float"
                    )

                if fps < 1:
                    if codec in IMG_CODECS:
                        fps = 25
                    else:
                        log.error("fps cannot be less than 1.")

                self.videos.append(
                    VideoStream(stream["width"], stream["height"], codec, fps, br, lang)
                )
            if codec_type == "audio":
                sr = int(stream["sample_rate"])
                self.audios.append(AudioStream(codec, sr, br, lang))
            if codec_type == "subtitle":
                ext = SUB_EXTS.get(codec, "vtt")
                self.subtitles.append(SubtitleStream(codec, ext, lang))



import sys

from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.utils.log import Log
from auto_editor.vanparse import ArgumentParser


def desc_options(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument("--ffmpeg-location", help="Point to your custom ffmpeg file.")
    parser.add_required("input", nargs="*", help="Path to file(s)")
    return parser


def main(sys_args=sys.argv[1:]) -> None:
    args = desc_options(ArgumentParser("desc")).parse_args(sys_args)
    for input_file in args.input:
        inp = FileInfo(input_file, FFmpeg(args.ffmpeg_location, debug=False), Log())
        if inp.description is not None:
            sys.stdout.write(f"\n{inp.description}\n\n")
        else:
            sys.stdout.write("\nNo description.\n\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
import sys
import tempfile

import auto_editor
import auto_editor.utils.func as usefulfunctions
from auto_editor.edit import edit_media
from auto_editor.ffwrapper import FFmpeg
from auto_editor.utils.log import Log, Timer
from auto_editor.validate_input import valid_input
from auto_editor.vanparse import ArgumentParser


def main_options(parser: ArgumentParser) -> ArgumentParser:
    from auto_editor.utils.types import (
        color_type,
        float_type,
        frame_type,
        margin_type,
        range_type,
        sample_rate_type,
        speed_range_type,
    )

    parser.add_text("Object Options")
    parser.add_argument(
        "--add-text",
        nargs="*",
        help="Add a text object to the timeline.",
    )
    parser.add_argument(
        "--add-rectangle",
        nargs="*",
        help="Add a rectangle object to the timeline.",
    )
    parser.add_argument(
        "--add-ellipse",
        nargs="*",
        help="Add an ellipse object to the timeline.",
    )
    parser.add_argument(
        "--add-image",
        nargs="*",
        help="Add an image object onto the timeline.",
    )
    parser.add_text("URL Download Options")
    parser.add_argument(
        "--yt-dlp-location", default="yt-dlp", help="Set a custom path to yt-dlp."
    )
    parser.add_argument(
        "--download-format", help="Set the yt-dlp download format. (--format, -f)"
    )
    parser.add_argument(
        "--output-format", help="Set the yt-dlp output file template. (--output, -o)"
    )
    parser.add_argument(
        "--yt-dlp-extras", help="Add extra options for yt-dlp. Must be in quotes"
    )
    parser.add_text("Exporting as Media Options")
    parser.add_argument(
        "--video-codec",
        "-vcodec",
        "-c:v",
        default="auto",
        help="Set the video codec for the output media file.",
    )
    parser.add_argument(
        "--audio-codec",
        "-acodec",
        "-c:a",
        default="auto",
        help="Set the audio codec for the output media file.",
    )
    parser.add_argument(
        "--video-bitrate",
        "-b:v",
        default="10m",
        help="Set the number of bits per second for video.",
    )
    parser.add_argument(
        "--audio-bitrate",
        "-b:a",
        default="unset",
        help="Set the number of bits per second for audio.",
    )
    parser.add_argument(
        "--video-quality-scale",
        "-qscale:v",
        "-q:v",
        default="unset",
        help="Set a value to the ffmpeg option -qscale:v",
    )
    parser.add_argument(
        "--scale",
        type=float_type,
        default=1,
        help="Scale the input video's resolution by the given factor.",
    )
    parser.add_argument(
        "--extras",
        help="Add extra options for ffmpeg for video rendering. Must be in quotes.",
    )
    parser.add_argument(
        "--no-seek",
        flag=True,
        help="Disable file seeking when rendering video. Helpful for debugging desync issues.",
    )
    parser.add_text("Manual Editing Options")
    parser.add_argument(
        "--cut-out",
        type=range_type,
        nargs="*",
        help="The range of media that will be removed completely, regardless of the "
        "value of silent speed.",
    )
    parser.add_argument(
        "--add-in",
        type=range_type,
        nargs="*",
        help="The range of media that will be added in, opposite of --cut-out",
    )
    parser.add_blank()
    parser.add_argument(
        "--mark-as-loud",
        type=range_type,
        nargs="*",
        help='The range that will be marked as "loud".',
    )
    parser.add_argument(
        "--mark-as-silent",
        type=range_type,
        nargs="*",
        help='The range that will be marked as "silent".',
    )
    parser.add_argument(
        "--set-speed-for-range",
        "--set-speed",
        type=speed_range_type,
        nargs="*",
        help="SPEED,START,STOP - Set an arbitrary speed for a given range.",
    )
    parser.add_text("Timeline Options")
    parser.add_argument(
        "--frame-rate",
        "-fps",
        "-r",
        type=float,
        help="Set the frame rate for the timeline and output media.",
    )
    parser.add_argument(
        "--sample-rate",
        "-ar",
        type=sample_rate_type,
        help="Set the sample rate for the timeline and output media.",
    )
    parser.add_argument(
        "--background",
        type=color_type,
        default="#000",
        help="Set the color of the background that is visible when the video is moved.",
    )
    parser.add_text("Select Editing Source Options")
    parser.add_argument(
        "--edit-based-on",
        "--edit",
        default="audio",
        help="Decide which method to use when making edits.",
    )
    parser.add_argument(
        "--keep-tracks-separate",
        flag=True,
        help="Don't combine audio tracks when exporting.",
    )
    parser.add_argument(
        "--export",
        "-ex",
        default="default",
        choices=[
            "default",
            "premiere",
            "final-cut-pro",
            "shotcut",
            "json",
            "audio",
            "clip-sequence",
        ],
        help="Choose the export mode.",
    )
    parser.add_text("Utility Options")
    parser.add_argument(
        "--no-open", flag=True, help="Do not open the file after editing is done."
    )
    parser.add_argument(
        "--temp-dir",
        help="Set where the temporary directory is located.",
    )
    parser.add_argument(
        "--ffmpeg-location",
        help="Set a custom path to the ffmpeg location.",
    )
    parser.add_argument(
        "--my-ffmpeg",
        flag=True,
        help="Use the ffmpeg on your PATH instead of the one packaged.",
    )
    parser.add_text("Display Options")
    parser.add_argument(
        "--progress",
        default="modern",
        choices=["modern", "classic", "ascii", "machine", "none"],
        help="Set what type of progress bar to use.",
    )
    parser.add_argument(
        "--version", flag=True, help="Display the program's version and halt."
    )
    parser.add_argument(
        "--debug", flag=True, help="Show debugging messages and values."
    )
    parser.add_argument(
        "--show-ffmpeg-debug", flag=True, help="Show ffmpeg progress and output."
    )
    parser.add_argument("--quiet", "-q", flag=True, help="Display less output.")
    parser.add_argument(
        "--preview", flag=True, help="Show stats on how the input will be cut and halt."
    )
    parser.add_argument(
        "--timeline", flag=True, help="Show auto-editor JSON timeline file and halt."
    )
    parser.add_argument(
        "--api",
        default="1.0.0",
        help="Set what version of the JSON timeline to output.",
    )
    parser.add_text("Global Editing Options")
    parser.add_argument(
        "--silent-threshold",
        "-t",
        type=float_type,
        default=0.04,
        help="Set the volume that frames audio needs to surpass to be marked loud.",
    )
    parser.add_argument(
        "--frame-margin",
        "--margin",
        "-m",
        type=margin_type,
        default="6",
        help='Set how many "silent" frames on either side of the "loud" sections to include.',
    )
    parser.add_argument(
        "--silent-speed",
        "-s",
        type=float_type,
        default=99999,
        help='Set the speed that "silent" sections should be played at.',
    )
    parser.add_argument(
        "--video-speed",
        "--sounded-speed",
        "-v",
        type=float_type,
        default=1,
        help='Set the speed that "loud" sections should be played at.',
    )
    parser.add_argument(
        "--min-clip-length",
        "-minclip",
        "-mclip",
        type=frame_type,
        default="3",
        help="Set the minimum length a clip can be. If a clip is too short, cut it.",
    )
    parser.add_argument(
        "--min-cut-length",
        "-mincut",
        "-mcut",
        type=frame_type,
        default="6",
        help="Set the minimum length a cut can be. If a cut is too short, don't cut.",
    )
    parser.add_blank()
    parser.add_argument(
        "--output-file",
        "--output",
        "-o",
        help="Set the name/path of the new output file.",
    )
    parser.add_required(
        "input", nargs="*", help="File(s) or URL(s) that will be edited."
    )

    return parser


def main() -> None:
    subcommands = ("test", "info", "levels", "grep", "subdump", "desc")

    if len(sys.argv) > 1 and sys.argv[1] in subcommands:
        obj = __import__(
            f"auto_editor.subcommands.{sys.argv[1]}", fromlist=["subcommands"]
        )
        obj.main(sys.argv[2:])
        sys.exit()
    else:
        args = main_options(ArgumentParser("Auto-Editor")).parse_args(
            sys.argv[1:],
            macros=[
                ({"--export-to-premiere", "-exp"}, ["--export", "premiere"]),
                ({"--export-to-final-cut-pro", "-exf"}, ["--export", "final-cut-pro"]),
                ({"--export-to-shotcut", "-exs"}, ["--export", "shotcut"]),
                ({"--export-as-json"}, ["--export", "json"]),
                ({"--export-as-clip-sequence", "-excs"}, ["--export", "clip-sequence"]),
                ({"--combine-files"}, []),
                ({"--keep-tracks-seperate"}, ["--keep-tracks-separate"]),
            ],
        )

    timer = Timer(args.quiet)

    exporting_to_editor = args.export in ("premiere", "final-cut-pro", "shotcut")

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, args.show_ffmpeg_debug)

    if args.debug and args.input == []:
        import platform as plat

        is64bit = "64-bit" if sys.maxsize > 2**32 else "32-bit"
        print(f"Python Version: {plat.python_version()} {is64bit}")
        print(f"Platform: {plat.system()} {plat.release()} {plat.machine().lower()}")
        print(f"FFmpeg Version: {ffmpeg.version}\nFFmpeg Path: {ffmpeg.path}")
        print(f"Auto-Editor Version: {auto_editor.version}")
        sys.exit()

    if args.version:
        print(f"{auto_editor.version} ({auto_editor.__version__})")
        sys.exit()

    if args.timeline:
        args.quiet = True

    if args.input == []:
        Log().error("You need to give auto-editor an input file.")

    if args.temp_dir is None:
        temp = tempfile.mkdtemp()
    else:
        temp = args.temp_dir
        if os.path.isfile(temp):
            Log().error("Temp directory cannot be an already existing file.")
        if os.path.isdir(temp):
            if len(os.listdir(temp)) != 0:
                Log().error("Temp directory should be empty!")
        else:
            os.mkdir(temp)

    log = Log(args.debug, args.quiet, temp=temp)
    log.debug(f"Temp Directory: {temp}")

    log.conwrite("Starting")

    if args.preview or args.export not in ("audio", "default"):
        args.no_open = True

    if args.silent_speed <= 0 or args.silent_speed > 99999:
        args.silent_speed = 99999

    if args.video_speed <= 0 or args.video_speed > 99999:
        args.video_speed = 99999

    inputs = valid_input(args.input, ffmpeg, args, log)

    if exporting_to_editor and len(inputs) > 1:
        cmd = []
        for inp in inputs:
            cmd.extend(["-i", inp])
        cmd.extend(
            [
                "-filter_complex",
                f"[0:v]concat=n={len(inputs)}:v=1:a=1",
                "-codec:v",
                "h264",
                "-pix_fmt",
                "yuv420p",
                "-strict",
                "-2",
                "combined.mp4",
            ]
        )
        ffmpeg.run(cmd)
        inputs = ["combined.mp4"]
    try:
        output = edit_media(inputs, ffmpeg, args, temp, log)

        if not args.preview and not args.timeline:
            timer.stop()

        if not args.no_open and output is not None:
            usefulfunctions.open_with_system_default(output, log)
    except KeyboardInterrupt:
        log.error("Keyboard Interrupt")
    log.cleanup()


if __name__ == "__main__":
    main()



from dataclasses import asdict, dataclass, fields
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

from auto_editor.ffwrapper import FileInfo
from auto_editor.method import get_speed_list
from auto_editor.objects import (
    AudioObj,
    EllipseObj,
    ImageObj,
    RectangleObj,
    TextObj,
    VideoObj,
)
from auto_editor.utils.func import chunkify, parse_dataclass
from auto_editor.utils.log import Log
from auto_editor.utils.progressbar import ProgressBar
from auto_editor.utils.types import (
    AlignType,
    Chunks,
    align_type,
    anchor_type,
    color_type,
    float_type,
)

Clip = NamedTuple(
    "Clip",
    [("start", int), ("dur", int), ("offset", int), ("speed", float), ("src", int)],
)

Visual = Union[TextObj, ImageObj, RectangleObj, EllipseObj]
VLayer = List[Union[VideoObj, Visual]]
VSpace = List[VLayer]

ALayer = List[AudioObj]
ASpace = List[ALayer]


def unclipify(layer: List[Clip]) -> NDArray[np.float_]:
    l: List[int] = []
    for clip in layer:
        if clip.src != 0:
            raise ValueError("Clip has src that is not 0")

        if clip.start > len(l):
            raise ValueError(
                f"Clip layer has null frames, cannot convert speed list: {l}"
            )

        for item in range(clip.offset, clip.offset + clip.dur):
            l.append(item)

    if len(set(l)) != len(l) or sorted(l) != l:
        raise ValueError(f"Clip layer to complex, cannot convert to speed list: {l}")

    arr = np.empty(layer[-1].offset + layer[-1].dur, dtype=float)
    arr.fill(99999)

    for clip in layer:
        arr[clip.offset : clip.offset + clip.dur] = clip.speed

    return arr


def _values(
    name: str,
    val: Union[float, str],
    _type: Union[type, Callable[[Any], Any]],
    _vars: Dict[str, int],
    log: Log,
):
    if _type is Any:
        return None
    if _type is float and name != "rotate":
        _type = float_type
    elif _type == AlignType:
        _type = align_type
    elif name == "anchor":
        _type = anchor_type
    elif name in ("fill", "strokecolor"):
        _type = color_type

    if _type is int:
        for key, item in _vars.items():
            if val == key:
                return item

    try:
        _type(val)
    except TypeError as e:
        log.error(str(e))
    except Exception:
        log.error(f"variable '{val}' is not defined.")

    return _type(val)


@dataclass
class Timeline:
    inputs: List[FileInfo]
    fps: float
    samplerate: int
    res: Tuple[int, int]
    background: str
    v: VSpace
    a: ASpace
    chunks: Optional[Chunks] = None

    @property
    def inp(self) -> FileInfo:
        return self.inputs[0]

    def out_len(self) -> float:
        out_len: float = 0
        for vclips in self.v:
            dur: float = 0
            for v_obj in vclips:
                if isinstance(v_obj, VideoObj):
                    dur += v_obj.dur / v_obj.speed
                else:
                    dur += v_obj.dur
            out_len = max(out_len, dur)
        for aclips in self.a:
            dur = 0
            for aclip in aclips:
                dur += aclip.dur / aclip.speed
            out_len = max(out_len, dur)
        return out_len


def clipify(chunks: Chunks, src: int) -> List[Clip]:
    clips = []
    start = 0
    # Add "+1" to match how chunks are rendered in 22w18a
    i = 0
    for chunk in chunks:
        if chunk[2] != 99999:
            if i == 0:
                dur = chunk[1] - chunk[0] + 1
                offset = chunk[0]
            else:
                dur = chunk[1] - chunk[0]
                offset = chunk[0] + 1

            clips.append(Clip(start, dur, offset, chunk[2], src))
            start += dur
            i += 1

    # If Chunks equals: [
    #    (0, 26, 1.0),
    #    (26, 34, 99999.0),
    #    (34, 396, 1.0),
    #    (396, 410, 99999.0),
    #    (410, 522, 1.0), (522, 1192, 99999.0),
    #    (1192, 1220, 1.0),
    #    (1220, 1273, 99999.0)
    # ]
    # try:
    #     assert clips == [
    #         Clip(start=0, dur=27, offset=0, speed=1.0, src=0),
    #         Clip(start=27, dur=362, offset=35, speed=1.0, src=0),
    #         Clip(start=27 + 362, dur=112, offset=411, speed=1.0, src=0),
    #         Clip(start=27 + 362 + 112, dur=28, offset=1193, speed=1.0, src=0),
    #     ]
    # except AssertionError:
    #     for clip in clips:
    #         print(clip)
    #     quit(1)

    return clips


def make_av(clips: List[Clip], inp: FileInfo) -> Tuple[VSpace, ASpace]:

    vclips: VSpace = [[]]
    aclips: ASpace = [[] for a in inp.audios]

    for clip in clips:
        vclips[0].append(
            VideoObj(clip.start, clip.dur, clip.offset, clip.speed, clip.src)
        )

    for a, _ in enumerate(inp.audios):
        for clip in clips:
            aclips[a].append(
                AudioObj(clip.start, clip.dur, clip.offset, clip.speed, clip.src, a)
            )

    return vclips, aclips


def make_layers(
    inputs: List[FileInfo], speedlists: List[NDArray[np.float_]]
) -> Tuple[int, Optional[Chunks], VSpace, ASpace]:

    clips = []
    for i, _chunks in enumerate([chunkify(s) for s in speedlists]):
        clips += clipify(_chunks, i)

    if len(clips) == 0:
        end = 0
    else:
        e = clips[-1]
        end = round(e.start + (e.dur * e.speed))

    chunks: Optional[Chunks] = None
    try:
        chunks = chunkify(unclipify(clips))
    except ValueError:
        pass

    vclips, aclips = make_av(clips, inputs[0])
    return end, chunks, vclips, aclips


def make_timeline(
    inputs: List[FileInfo], args, sr: int, progress: ProgressBar, temp: str, log: Log
) -> Timeline:
    assert len(inputs) > 0

    inp = inputs[0]

    if args.frame_rate is None:
        fps = inp.get_fps()
    else:
        fps = args.frame_rate

    res = inp.get_res()

    speedlists = []
    for i, inp in enumerate(inputs):
        speedlists.append(get_speed_list(i, inp, fps, args, progress, temp, log))

    end, chunks, vclips, aclips = make_layers(inputs, speedlists)

    timeline = Timeline(inputs, fps, sr, res, args.background, vclips, aclips, chunks)

    w, h = res
    _vars: Dict[str, int] = {
        "width": w,
        "height": h,
        "centerX": w // 2,
        "centerY": h // 2,
        "start": 0,
        "end": end,
    }

    pool = []

    for o in args.add_text:
        pool.append(parse_dataclass(o, TextObj, log))
    for o in args.add_rectangle:
        pool.append(parse_dataclass(o, RectangleObj, log))
    for o in args.add_ellipse:
        pool.append(parse_dataclass(o, EllipseObj, log))
    for o in args.add_image:
        pool.append(parse_dataclass(o, ImageObj, log))

    for obj in pool:
        dic_value = asdict(obj)
        dic_type = {}
        for field in fields(obj):
            dic_type[field.name] = field.type

        # Convert to the correct types
        for k, _type in dic_type.items():
            obj.__setattr__(k, _values(k, dic_value[k], _type, _vars, log))

        if obj.dur < 1:
            log.error(f"dur's value must be greater than 0. Was '{obj.dur}'.")

    # Higher layers are visually on top
    for obj in pool:
        timeline.v.append([obj])

    return timeline



import os
import sys
import tempfile

from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.utils.log import Log
from auto_editor.vanparse import ArgumentParser


def subdump_options(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument("--ffmpeg-location", help="Point to your custom ffmpeg file.")
    parser.add_argument(
        "--my-ffmpeg",
        flag=True,
        help="Use the ffmpeg on your PATH instead of the one packaged.",
    )
    parser.add_required(
        "input", nargs="*", help="Path to the file to have its subtitles dumped."
    )
    return parser


def main(sys_args=sys.argv[1:]) -> None:
    parser = subdump_options(ArgumentParser("subdump"))
    args = parser.parse_args(sys_args)

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, debug=False)

    temp = tempfile.mkdtemp()
    log = Log(temp=temp)

    for i, input_file in enumerate(args.input):
        inp = FileInfo(input_file, ffmpeg, log)

        cmd = ["-i", input_file]
        for s, sub in enumerate(inp.subtitles):
            cmd.extend(["-map", f"0:s:{s}", os.path.join(temp, f"{i}s{s}.{sub.ext}")])
        ffmpeg.run(cmd)

        for s, sub in enumerate(inp.subtitles):
            print(f"file: {input_file} ({s}:{sub.lang}:{sub.ext})")
            with open(os.path.join(temp, f"{i}s{s}.{sub.ext}")) as file:
                print(file.read())
            print("------")

    log.cleanup()


if __name__ == "__main__":
    main()



import json
import os.path
import sys

import av

from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.utils.func import aspect_ratio
from auto_editor.utils.log import Log
from auto_editor.vanparse import ArgumentParser

av.logging.set_level(av.logging.PANIC)


def info_options(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument("--json", flag=True, help="Export info in JSON format.")
    parser.add_argument(
        "--include-vfr",
        "--has-vfr",
        flag=True,
        help="Display the number of Variable Frame Rate (VFR) frames.",
    )
    parser.add_argument("--ffmpeg-location", help="Point to your custom ffmpeg file.")
    parser.add_argument(
        "--my-ffmpeg",
        flag=True,
        help="Use the ffmpeg on your PATH instead of the one packaged.",
    )
    parser.add_required(
        "input", nargs="*", help="The path to a file you want inspected."
    )
    return parser


def main(sys_args=sys.argv[1:]):
    parser = info_options(ArgumentParser("info"))
    args = parser.parse_args(sys_args)

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, False)

    file_info = {}

    for file in args.input:
        if not os.path.isfile(file):
            Log().error(f"Could not find file: {file}")

        inp = FileInfo(file, ffmpeg, Log())

        if len(inp.videos) + len(inp.audios) + len(inp.subtitles) == 0:
            file_info[file] = {"media": "invalid"}
            continue

        file_info[file] = {
            "video": [],
            "audio": [],
            "subtitle": [],
            "container": {},
        }

        cn = av.open(file, "r")

        for track, stream in enumerate(inp.videos):
            pix_fmt = cn.streams.video[track].pix_fmt
            time_base = str(cn.streams.video[track].time_base)
            cc_time_base = str(cn.streams.video[track].codec_context.time_base)
            w, h = stream.width, stream.height
            w_, h_ = aspect_ratio(w, h)

            fps = stream.fps
            if fps is not None and int(fps) == float(fps):
                fps = int(fps)

            vid = {
                "codec": stream.codec,
                "pix_fmt": pix_fmt,
                "fps": fps,
                "resolution": [w, h],
                "aspect ratio": [w_, h_],
                "timebase": time_base,
                "cc timebase": cc_time_base,
                "bitrate": stream.bitrate,
                "lang": stream.lang,
            }
            file_info[file]["video"].append(vid)

        for track, stream in enumerate(inp.audios):
            aud = {
                "codec": stream.codec,
                "samplerate": stream.samplerate,
                "bitrate": stream.bitrate,
                "lang": stream.lang,
            }
            file_info[file]["audio"].append(aud)

        for track, stream in enumerate(inp.subtitles):
            sub = {"codec": stream.codec, "lang": stream.lang}
            file_info[file]["subtitle"].append(sub)

        cont = {"bitrate": inp.bitrate}

        if args.include_vfr:
            fps_mode = ffmpeg.pipe(
                [
                    "-i",
                    file,
                    "-hide_banner",
                    "-vf",
                    "vfrdet",
                    "-an",
                    "-f",
                    "null",
                    "-",
                ]
            ).strip()
            if "VFR:" in fps_mode:
                fps_mode = (fps_mode[fps_mode.index("VFR:") :]).strip()

            cont["fps_mode"] = fps_mode

        file_info[file]["container"] = cont

    if args.json:
        print(json.dumps(file_info, indent=4))
        return

    def stream_to_text(text: str, label: str, streams) -> str:
        if len(streams) > 0:
            text += f" - {label}:\n"

        for s, stream in enumerate(streams):
            text += f"   - track {s}:\n"
            for key, value in stream.items():
                if value is not None:
                    if isinstance(value, list):
                        sep = "x" if key == "resolution" else ":"
                        value = sep.join([str(x) for x in value])

                    text += f"     - {key}: {value}\n"
        return text

    text = ""
    for name, info in file_info.items():
        text += f"{name}:\n"
        if "media" in info:
            text += " - invalid media\n\n"
            continue

        for label, streams in info.items():
            if isinstance(streams, dict):
                text += " - container:\n"
                for key, value in streams.items():
                    text += f"   - {key}: {value}\n"
            else:
                text = stream_to_text(text, label, streams)
        text += "\n"

    sys.stdout.write(text)


if __name__ == "__main__":
    main()



import numpy as np

from .array import ArrReader, ArrWriter


class CBuffer:
    def __init__(self, channels: int, max_length: int) -> None:
        self._data = np.zeros((channels, max_length), dtype=np.float32)
        self._channels = channels
        self._max_length = max_length

        self._offset = 0
        self._ready = 0
        self.length = 0

    def add(self, buffer: np.ndarray) -> None:
        """Adds a buffer element-wise to the CBuffer."""
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of channels")

        n = buffer.shape[1]
        if n > self.length:
            raise ValueError("not enough space remaining in CBuffer")

        # Compute the slice of data where the values will be added
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] += buffer[:, :n]
        else:
            end -= self._max_length
            self._data[:, start:] += buffer[:, : self._max_length - start]
            self._data[:, :end] += buffer[:, self._max_length - start : n]

    def divide(self, array: np.ndarray) -> None:
        n = len(array)
        if n > self.length:
            raise ValueError("not enough space remaining in the CBuffer")

        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] /= array[:n]
        else:
            end -= self._max_length
            self._data[:, start:] /= array[: self._max_length - start]
            self._data[:, :end] /= array[self._max_length - start : n]

    def peek(self, buffer: np.ndarray) -> int:
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of channels")

        n = min(buffer.shape[1], self._ready)

        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            np.copyto(buffer[:, :n], self._data[:, start:end])
        else:
            end -= self._max_length
            np.copyto(buffer[:, : self._max_length - start], self._data[:, start:])
            np.copyto(buffer[:, self._max_length - start : n], self._data[:, :end])

        return n

    def read(self, buffer: np.ndarray) -> int:
        n = self.peek(buffer)
        self.remove(n)
        return n

    def read_from(self, reader: ArrReader) -> int:
        # Compute the slice of data that will be written to
        start = (self._offset + self.length) % self._max_length
        end = start + self._max_length - self.length

        if end <= self._max_length:
            n = reader.read(self._data[:, start:end])
        else:
            # There is not enough space to copy the whole buffer, it has to be
            # split into two parts, one of which will be copied at the end of
            # _data, and the other at the beginning.
            end -= self._max_length

            n = reader.read(self._data[:, start:])
            n += reader.read(self._data[:, :end])

        self.length += n
        self._ready = self.length
        return n

    @property
    def ready(self):
        return self._ready

    @property
    def remaining_length(self):
        return self._max_length - self._ready

    def remove(self, n: int) -> int:
        """
        Removes the first n samples of the CBuffer, preventing
        them to be read again, and leaving more space for new samples to be
        written.
        """
        if n >= self.length:
            n = self.length

        # Compute the slice of data that will be reset to 0
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] = 0
        else:
            end -= self._max_length
            self._data[:, start:] = 0
            self._data[:, :end] = 0

        self._offset += n
        self._offset %= self._max_length
        self.length -= n

        self._ready -= n
        if self._ready < 0:
            self._ready = 0

        return n

    def right_pad(self, n: int) -> None:
        if n > self._max_length - self.length:
            raise ValueError("not enough space remaining in CBuffer")

        self.length += n

    def set_ready(self, n: int) -> None:
        """Mark the next n samples as ready to be read."""
        if self._ready + n > self.length:
            raise ValueError("not enough samples to be marked as ready")

        self._ready += n

    def to_array(self):
        out = np.empty((self._channels, self._ready))
        self.peek(out)
        return out

    def write(self, buffer: np.ndarray) -> int:
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of channels")

        n = min(buffer.shape[1], self._max_length - self.length)

        # Compute the slice of data that will be written to
        start = (self._offset + self.length) % self._max_length
        end = start + n

        if end <= self._max_length:
            np.copyto(self._data[:, start:end], buffer[:, :n])
        else:
            # There is not enough space to copy the whole buffer, it has to be
            # split into two parts, one of which will be copied at the end of
            # _data, and the other at the beginning.
            end -= self._max_length

            np.copyto(self._data[:, start:], buffer[:, : self._max_length - start])
            np.copyto(self._data[:, :end], buffer[:, self._max_length - start : n])

        self.length += n
        self._ready = self.length
        return n

    def write_to(self, writer: ArrWriter) -> int:
        start = self._offset
        end = self._offset + self._ready

        if end <= self._max_length:
            n = writer.write(self._data[:, start:end])
        else:
            end -= self._max_length
            n = writer.write(self._data[:, start:])
            n += writer.write(self._data[:, :end])

        self.remove(n)
        return n



from typing import Tuple

import numpy as np
from numpy.typing import NDArray

from .array import ArrReader, ArrWriter
from .cbuffer import CBuffer
from .normalizebuffer import NormalizeBuffer

EPSILON = 0.0001


def find_peaks(amplitude: NDArray[np.float_]) -> NDArray[np.bool_]:
    # To avoid overflows
    padded = np.concatenate((-np.ones(2), amplitude, -np.ones(2)))

    # Shift the array by one/two values to the left/right
    shifted_l2 = padded[:-4]
    shifted_l1 = padded[1:-3]
    shifted_r1 = padded[3:-1]
    shifted_r2 = padded[4:]

    # Compare the original array with the shifted versions.
    peaks = (
        (amplitude >= shifted_l2)
        & (amplitude >= shifted_l1)
        & (amplitude >= shifted_r1)
        & (amplitude >= shifted_r2)
    )

    return peaks


def get_closest_peaks(peaks: NDArray[np.bool_]) -> NDArray[np.int_]:
    """
    Returns an array containing the index of the closest peak of each index.
    """
    closest_peak = np.empty_like(peaks, dtype=int)
    previous = -1
    for i, is_peak in enumerate(peaks):
        if is_peak:
            if previous >= 0:
                closest_peak[previous : (previous + i) // 2 + 1] = previous
                closest_peak[(previous + i) // 2 + 1 : i] = i
            else:
                closest_peak[:i] = i
            previous = i
    closest_peak[previous:] = previous

    return closest_peak


class PhaseVocoderConverter:
    def __init__(
        self, channels: int, frame_length: int, analysis_hop: int, synthesis_hop: int
    ) -> None:
        self.channels = channels
        self._frame_length = frame_length
        self._synthesis_hop = synthesis_hop
        self._analysis_hop = analysis_hop

        self._center_frequency = np.fft.rfftfreq(frame_length) * 2 * np.pi  # type: ignore
        fft_length = len(self._center_frequency)

        self._first = True

        self._previous_phase = np.empty((channels, fft_length))
        self._output_phase = np.empty((channels, fft_length))

        # Buffer used to compute the phase increment and the instantaneous frequency
        self._buffer = np.empty(fft_length)

    def clear(self) -> None:
        self._first = True

    def convert_frame(self, frame: np.ndarray) -> np.ndarray:
        for k in range(self.channels):
            # Compute the FFT of the analysis frame
            stft = np.fft.rfft(frame[k])
            amplitude = np.abs(stft)

            phase: NDArray[np.float_]
            phase = np.angle(stft)  # type: ignore
            del stft

            peaks = find_peaks(amplitude)
            closest_peak = get_closest_peaks(peaks)

            if self._first:
                # Leave the first frame unchanged
                self._output_phase[k, :] = phase
            else:
                # Compute the phase increment
                self._buffer[peaks] = (
                    phase[peaks]
                    - self._previous_phase[k, peaks]
                    - self._analysis_hop * self._center_frequency[peaks]
                )

                # Unwrap the phase increment
                self._buffer[peaks] += np.pi
                self._buffer[peaks] %= 2 * np.pi
                self._buffer[peaks] -= np.pi

                # Compute the instantaneous frequency (in the same buffer,
                # since the phase increment wont be required after that)
                self._buffer[peaks] /= self._analysis_hop
                self._buffer[peaks] += self._center_frequency[peaks]

                self._buffer[peaks] *= self._synthesis_hop
                self._output_phase[k][peaks] += self._buffer[peaks]

                # Phase locking
                self._output_phase[k] = (
                    self._output_phase[k][closest_peak] + phase - phase[closest_peak]
                )

                # Compute the new stft
                output_stft = amplitude * np.exp(1j * self._output_phase[k])

                frame[k, :] = np.fft.irfft(output_stft).real

            # Save the phase for the next analysis frame
            self._previous_phase[k, :] = phase
            del phase
            del amplitude

        self._first = False

        return frame


class AnalysisSynthesisTSM:
    def run(self, reader: ArrReader, writer: ArrWriter, flush: bool = True) -> None:
        finished = False
        while not (finished and reader.empty):
            self.read_from(reader)
            _, finished = self.write_to(writer)

        if flush:
            finished = False
            while not finished:
                _, finished = self.flush_to(writer)

            self.clear()

    def __init__(
        self,
        channels: int,
        frame_length: int,
        analysis_hop: int,
        synthesis_hop: int,
        analysis_window: np.ndarray,
        synthesis_window: np.ndarray,
    ) -> None:

        self._converter = PhaseVocoderConverter(
            channels, frame_length, analysis_hop, synthesis_hop
        )

        self._channels = channels
        self._frame_length = frame_length
        self._analysis_hop = analysis_hop
        self._synthesis_hop = synthesis_hop

        self._analysis_window = analysis_window
        self._synthesis_window = synthesis_window

        # When the analysis hop is larger than the frame length, some samples
        # from the input need to be skipped.
        self._skip_input_samples = 0

        # Used to start the output signal in the middle of a frame, which should
        # be the peek of the window function
        self._skip_output_samples = 0

        self._normalize_window = self._analysis_window * self._synthesis_window

        # Initialize the buffers
        self._in_buffer = CBuffer(self._channels, self._frame_length)
        self._analysis_frame = np.empty((self._channels, self._frame_length))
        self._out_buffer = CBuffer(self._channels, self._frame_length)
        self._normalize_buffer = NormalizeBuffer(self._frame_length)

        self.clear()

    def clear(self) -> None:
        self._in_buffer.remove(self._in_buffer.length)
        self._out_buffer.remove(self._out_buffer.length)
        self._out_buffer.right_pad(self._frame_length)
        self._normalize_buffer.remove(self._normalize_buffer.length)

        # Left pad the input with half a frame of zeros, and ignore that half
        # frame in the output. This makes the output signal start in the middle
        # of a frame, which should be the peak of the window function.
        self._in_buffer.write(np.zeros((self._channels, self._frame_length // 2)))
        self._skip_output_samples = self._frame_length // 2

        self._converter.clear()

    def flush_to(self, writer: ArrWriter) -> Tuple[int, bool]:
        if self._in_buffer.remaining_length == 0:
            raise RuntimeError(
                "There is still data to process in the input buffer, flush_to method "
                "should only be called when write_to returns True."
            )

        n = self._out_buffer.write_to(writer)
        if self._out_buffer.ready == 0:
            # The output buffer is empty
            self.clear()
            return n, True

        return n, False

    def get_max_output_length(self, input_length: int) -> int:
        input_length -= self._skip_input_samples
        if input_length <= 0:
            return 0

        n_frames = input_length // self._analysis_hop + 1
        return n_frames * self._synthesis_hop

    def _process_frame(self) -> None:
        """Read an analysis frame from the input buffer, process it, and write
        the result to the output buffer."""
        # Generate the analysis frame and discard the input samples that will
        # not be needed anymore
        self._in_buffer.peek(self._analysis_frame)
        self._in_buffer.remove(self._analysis_hop)

        for channel in self._analysis_frame:
            channel *= self._analysis_window

        synthesis_frame = self._converter.convert_frame(self._analysis_frame)

        for channel in synthesis_frame:
            channel *= self._synthesis_window

        # Overlap and add the synthesis frame in the output buffer
        self._out_buffer.add(synthesis_frame)

        # The overlap and add step changes the volume of the signal. The
        # normalize_buffer is used to keep track of "how much of the input
        # signal was added" to each part of the output buffer, allowing to
        # normalize it.
        self._normalize_buffer.add(self._normalize_window)

        # Normalize the samples that are ready to be written to the output
        normalize = self._normalize_buffer.to_array(end=self._synthesis_hop)
        normalize[normalize < EPSILON] = 1
        self._out_buffer.divide(normalize)
        self._out_buffer.set_ready(self._synthesis_hop)
        self._normalize_buffer.remove(self._synthesis_hop)

    def read_from(self, reader: ArrReader) -> int:
        n = reader.skip(self._skip_input_samples)
        self._skip_input_samples -= n
        if self._skip_input_samples > 0:
            return n

        n += self._in_buffer.read_from(reader)

        if (
            self._in_buffer.remaining_length == 0
            and self._out_buffer.remaining_length >= self._synthesis_hop
        ):
            # The input buffer has enough data to process, and there is enough
            # space in the output buffer to store the output
            self._process_frame()

            # Skip output samples if necessary
            skipped = self._out_buffer.remove(self._skip_output_samples)
            self._out_buffer.right_pad(skipped)
            self._skip_output_samples -= skipped

            # Set the number of input samples to be skipped
            self._skip_input_samples = self._analysis_hop - self._frame_length
            if self._skip_input_samples < 0:
                self._skip_input_samples = 0

        return n

    def write_to(self, writer: ArrWriter) -> Tuple[int, bool]:
        n = self._out_buffer.write_to(writer)
        self._out_buffer.right_pad(n)

        if self._in_buffer.remaining_length > 0 and self._out_buffer.ready == 0:
            # There is not enough data to process in the input buffer, and the
            # output buffer is empty
            return n, True

        return n, False



import os
import random
from dataclasses import asdict, dataclass, fields
from math import ceil
from typing import Any, Callable, Dict, Type, TypeVar, Union

import numpy as np
from numpy.typing import NDArray

from auto_editor.ffwrapper import FileInfo
from auto_editor.utils.func import (
    apply_margin,
    apply_mark_as,
    cook,
    parse_dataclass,
    seconds_to_frames,
    set_range,
    to_speed_list,
)
from auto_editor.utils.log import Log
from auto_editor.utils.progressbar import ProgressBar
from auto_editor.utils.types import StreamType, float_type, stream_type
from auto_editor.wavfile import read

T = TypeVar("T")


def get_attributes(attrs_str: str, dataclass: T, log: Log) -> T:
    attrs: T = parse_dataclass(attrs_str, dataclass, log)

    dic_value = asdict(attrs)
    dic_type: Dict[str, Union[type, Callable[[Any], Any]]] = {}
    for field in fields(attrs):
        dic_type[field.name] = field.type

    # Convert to the correct types
    for k, _type in dic_type.items():

        if _type == float:
            _type = float_type
        elif _type == StreamType:
            _type = stream_type

        try:
            attrs.__setattr__(k, _type(dic_value[k]))
        except (ValueError, TypeError) as e:
            log.error(str(e))

    return attrs


def get_media_duration(path: str, fps: float, temp: str, log: Log) -> int:
    audio_path = os.path.join(temp, "0-0.wav")
    if os.path.isfile(audio_path):
        sample_rate, audio_samples = read(audio_path)
        sample_rate_per_frame = sample_rate / fps
        return round(audio_samples.shape[0] / sample_rate_per_frame)

    import av

    cn = av.open(path, "r")

    if len(cn.streams.video) < 1:
        log.error("Could not get media duration")

    video = cn.streams.video[0]
    return int(float(video.duration * video.time_base) * fps)


def get_audio_list(
    i: int,
    stream: int,
    threshold: float,
    fps: float,
    progress: ProgressBar,
    temp: str,
    log: Log,
) -> NDArray[np.bool_]:

    from auto_editor.analyze.audio import audio_detection

    path = os.path.join(temp, f"{i}-{stream}.wav")

    if os.path.isfile(path):
        sample_rate, audio_samples = read(path)
    else:
        log.error(f"Audio stream '{stream}' does not exist.")

    audio_list = audio_detection(audio_samples, sample_rate, fps, progress)

    return np.fromiter((x > threshold for x in audio_list), dtype=np.bool_)


def operand_combine(
    a: NDArray[np.bool_], b: NDArray[np.bool_], call: Callable
) -> NDArray[np.bool_]:
    if len(a) > len(b):
        b = np.resize(b, len(a))
    if len(b) > len(a):
        a = np.resize(a, len(b))

    return call(a, b)


def get_stream_data(
    method: str,
    attrs,
    args,
    i: int,
    inp: FileInfo,
    fps: float,
    progress: ProgressBar,
    temp: str,
    log: Log,
) -> NDArray[np.bool_]:

    if method == "none":
        return np.ones(
            (get_media_duration(inp.path, fps, temp, log) - 1), dtype=np.bool_
        )
    if method == "all":
        return np.zeros(
            (get_media_duration(inp.path, fps, temp, log) - 1), dtype=np.bool_
        )
    if method == "random":
        if attrs.cutchance > 1 or attrs.cutchance < 0:
            log.error(f"random:cutchance must be between 0 and 1")
        l = get_media_duration(inp.path, fps, temp, log) - 1

        random.seed(attrs.seed)
        log.debug(f"Seed: {attrs.seed}")

        a = random.choices((0, 1), weights=(attrs.cutchance, 1 - attrs.cutchance), k=l)

        return np.asarray(a, dtype=np.bool_)
    if method == "audio":
        if attrs.stream == "all":
            total_list = None
            for s in range(len(inp.audios)):
                audio_list = get_audio_list(
                    i, s, attrs.threshold, fps, progress, temp, log
                )
                if total_list is None:
                    total_list = audio_list
                else:
                    total_list = operand_combine(total_list, audio_list, np.logical_or)

            if total_list is None:
                log.error("Input has no audio streams.")
            return total_list
        else:
            return get_audio_list(
                i, attrs.stream, attrs.threshold, fps, progress, temp, log
            )
    if method == "motion":
        from auto_editor.analyze.motion import motion_detection

        if len(inp.videos) == 0:
            log.error("Video stream '0' does not exist.")

        motion_list = motion_detection(inp.path, fps, progress, attrs.width, attrs.blur)
        return np.fromiter((x >= attrs.threshold for x in motion_list), dtype=np.bool_)

    # "pixeldiff"
    if method == "pixeldiff":
        from auto_editor.analyze.pixeldiff import pixel_difference

        if len(inp.videos) == 0:
            log.error("Video stream '0' does not exist.")

        pixel_list = pixel_difference(inp.path, fps, progress)
        return np.fromiter((x >= attrs.threshold for x in pixel_list), dtype=np.bool_)

    raise ValueError(f"Unreachable. {method=}")


def get_has_loud(
    method_str: str,
    i: int,
    inp: FileInfo,
    fps: float,
    progress: ProgressBar,
    temp: str,
    log: Log,
    args,
) -> NDArray[np.bool_]:
    @dataclass
    class Audio:
        stream: StreamType = 0
        threshold: float = args.silent_threshold

    @dataclass
    class Motion:
        threshold: float = 0.02
        blur: int = 9
        width: int = 400

    @dataclass
    class Pixeldiff:
        threshold: int = 1

    @dataclass
    class Random:
        cutchance: float = 0.5
        seed: int = random.randint(0, 2147483647)

    KEYWORD_SEP = " "
    METHOD_ATTRS_SEP = ":"

    result_array = None
    operand = None

    logic_funcs = {
        "and": np.logical_and,
        "or": np.logical_or,
        "xor": np.logical_xor,
    }

    Methods = Union[Type[Audio], Type[Motion], Type[Pixeldiff], Type[Random], None]

    method_str = method_str.replace("_", " ")  # Allow old style `--edit` to work

    for method in method_str.split(KEYWORD_SEP):

        if method == "":  # Skip whitespace
            continue

        if METHOD_ATTRS_SEP in method:
            method, attrs_str = method.split(METHOD_ATTRS_SEP)
        else:
            attrs_str = ""

        if method in ("audio", "motion", "pixeldiff", "random", "none", "all"):
            if result_array is not None and operand is None:
                log.error("Logic operator must be between two editing methods.")

            if method == "audio":
                attrs: Methods = get_attributes(attrs_str, Audio, log)
            elif method == "motion":
                attrs = get_attributes(attrs_str, Motion, log)
            elif method == "pixeldiff":
                attrs = get_attributes(attrs_str, Pixeldiff, log)
            elif method == "random":
                attrs = get_attributes(attrs_str, Random, log)
            else:
                attrs = None

            stream_data = get_stream_data(
                method, attrs, args, i, inp, fps, progress, temp, log
            )

            if operand == "not":
                result_array = np.logical_not(stream_data)
                operand = None
            elif result_array is None:
                result_array = stream_data
            elif operand is not None and operand in ("and", "or", "xor"):
                result_array = operand_combine(
                    result_array, stream_data, logic_funcs[operand]
                )
                operand = None

        elif method in ("and", "or", "xor"):
            if operand is not None:
                log.error("Invalid Editing Syntax.")
            if result_array is None:
                log.error(f"'{method}' operand needs two arguments.")
            operand = method
        elif method == "not":
            if operand is not None:
                log.error("Invalid Editing Syntax.")
            operand = method
        else:
            log.error(f"Unknown method/operator: '{method}'")

    if operand is not None:
        log.error(f"Dangling operand: '{operand}'")

    assert result_array is not None
    return result_array


def get_speed_list(
    i: int, inp: FileInfo, fps: float, args, progress: ProgressBar, temp: str, log: Log
) -> NDArray[np.float_]:

    start_margin, end_margin = args.frame_margin

    start_margin = seconds_to_frames(start_margin, fps)
    end_margin = seconds_to_frames(end_margin, fps)
    min_clip = seconds_to_frames(args.min_clip_length, fps)
    min_cut = seconds_to_frames(args.min_cut_length, fps)

    has_loud = get_has_loud(args.edit_based_on, i, inp, fps, progress, temp, log, args)
    has_loud_length = len(has_loud)

    has_loud = apply_mark_as(has_loud, has_loud_length, fps, args, log)
    has_loud = cook(has_loud, min_clip, min_cut)
    has_loud = apply_margin(has_loud, has_loud_length, start_margin, end_margin)

    # Remove small clips/cuts created by applying other rules.
    has_loud = cook(has_loud, min_clip, min_cut)

    speed_list = to_speed_list(has_loud, args.video_speed, args.silent_speed)

    if args.cut_out != []:
        speed_list = set_range(speed_list, args.cut_out, fps, 99999, log)

    if args.add_in != []:
        speed_list = set_range(speed_list, args.add_in, fps, args.video_speed, log)

    if args.set_speed_for_range != []:
        for item in args.set_speed_for_range:
            speed_list = set_range(speed_list, [item[1:]], fps, item[0], log)

    return speed_list



import os
import re
from dataclasses import dataclass
from typing import List, Tuple

from auto_editor.ffwrapper import FFmpeg
from auto_editor.timeline import Timeline
from auto_editor.utils.func import to_timecode
from auto_editor.utils.log import Log


@dataclass
class SerialSub:
    start: int
    end: int
    before: str
    middle: str
    after: str


class SubtitleParser:
    def __init__(self) -> None:
        self.supported_codecs = ("ass", "webvtt", "mov_text")

    def parse(self, text, fps: float, codec: str) -> None:

        if codec not in self.supported_codecs:
            raise ValueError(f"codec {codec} not supported.")

        self.fps = fps
        self.codec = codec
        self.contents: List[SerialSub] = []

        if codec == "ass":
            time_code = re.compile(r"(.*)(\d+:\d+:[\d.]+)(.*)(\d+:\d+:[\d.]+)(.*)")
        if codec == "webvtt":
            time_code = re.compile(r"()(\d+:[\d.]+)( --> )(\d+:[\d.]+)(\n.*)")
        if codec == "mov_text":
            time_code = re.compile(r"()(\d+:\d+:[\d,]+)( --> )(\d+:\d+:[\d,]+)(\n.*)")

        i = 0
        for reg in re.finditer(time_code, text):
            i += 1
            if i == 1:
                self.header = text[: reg.span()[0]]

            self.contents.append(
                SerialSub(
                    self.to_frame(reg.group(2)),
                    self.to_frame(reg.group(4)),
                    reg.group(1),
                    reg.group(3),
                    f"{reg.group(5)}\n",
                )
            )

        if i == 0:
            self.header = ""
            self.footer = ""
        else:
            self.footer = text[reg.span()[1] :]

    def edit(self, chunks: List[Tuple[int, int, float]]) -> None:
        for cut in reversed(chunks):
            the_speed = cut[2]
            speed_factor = 1 if the_speed == 99999 else 1 - (1 / the_speed)

            new_content = []
            for content in self.contents:
                if cut[0] <= content.end and cut[1] > content.start:

                    diff = int(
                        (min(cut[1], content.end) - max(cut[0], content.start))
                        * speed_factor
                    )
                    if content.start > cut[0]:
                        content.start -= diff
                        content.end -= diff

                    content.end -= diff

                elif content.start >= cut[0]:
                    diff = int((cut[1] - cut[0]) * speed_factor)

                    content.start -= diff
                    content.end -= diff

                if content.start != content.end:
                    new_content.append(content)

        self.contents = new_content

    def write(self, file_path: str) -> None:
        with open(file_path, "w") as file:
            file.write(self.header)
            for c in self.contents:
                file.write(
                    f"{c.before}{to_timecode(c.start / self.fps, self.codec)}"
                    f"{c.middle}{to_timecode(c.end / self.fps, self.codec)}"
                    f"{c.after}"
                )
            file.write(self.footer)

    def to_frame(self, text: str) -> int:
        if self.codec == "mov_text":
            time_format = r"(\d+):?(\d+):([\d,]+)"
        else:
            time_format = r"(\d+):?(\d+):([\d.]+)"

        nums = re.match(time_format, text)
        assert nums is not None

        hours, minutes, seconds = nums.groups()
        seconds = seconds.replace(",", ".", 1)
        return round(
            (int(hours) * 3600 + int(minutes) * 60 + float(seconds)) * self.fps
        )


def cut_subtitles(
    ffmpeg: FFmpeg,
    timeline: Timeline,
    temp: str,
    log: Log,
) -> None:
    inp = timeline.inp
    chunks = timeline.chunks

    for s, sub in enumerate(inp.subtitles):
        if chunks is None:
            log.error("Timeline too complex for subtitles")

        file_path = os.path.join(temp, f"{s}s.{sub.ext}")
        new_path = os.path.join(temp, f"new{s}s.{sub.ext}")

        parser = SubtitleParser()

        if sub.codec in parser.supported_codecs:
            with open(file_path) as file:
                parser.parse(file.read(), timeline.fps, sub.codec)
        else:
            convert_path = os.path.join(temp, f"{s}s_convert.vtt")
            ffmpeg.run(["-i", file_path, convert_path])
            with open(convert_path) as file:
                parser.parse(file.read(), timeline.fps, "webvtt")

        parser.edit(chunks)
        parser.write(new_path)



from typing import Tuple

import av
import numpy as np
from numpy.typing import NDArray
from PIL import ImageChops, ImageFilter, ImageOps

from auto_editor.utils.progressbar import ProgressBar


def new_size(size: Tuple[int, int], width: int) -> Tuple[int, int]:
    h, w = size
    return width, int(h * (width / w))


def motion_detection(
    path: str, fps: float, progress: ProgressBar, width: int, blur: int
) -> NDArray[np.float_]:
    container = av.open(path, "r")

    stream = container.streams.video[0]
    stream.thread_type = "AUTO"

    inaccurate_dur = int(stream.duration * stream.time_base * stream.rate)

    progress.start(inaccurate_dur, "Analyzing motion")

    prev_image = None
    image = None
    total_pixels = None
    index = 0

    threshold_list = np.zeros((1024), dtype=np.float_)

    for frame in container.decode(stream):
        if image is None:
            prev_image = None
        else:
            prev_image = image

        index = int(frame.time * fps)

        progress.tick(index)

        if index > len(threshold_list) - 1:
            threshold_list = np.concatenate(
                (threshold_list, np.zeros((len(threshold_list)), dtype=np.float_)),
                axis=0,
            )

        image = frame.to_image()

        if total_pixels is None:
            total_pixels = image.size[0] * image.size[1]

        image.thumbnail(new_size(image.size, width))
        image = ImageOps.grayscale(image)

        if blur > 0:
            image = image.filter(ImageFilter.GaussianBlur(radius=blur))

        if prev_image is not None:
            count = np.count_nonzero(ImageChops.difference(prev_image, image))

            threshold_list[index] = count / total_pixels

    progress.end()
    return threshold_list[:index]



from statistics import fmean, median
from typing import List, Optional, Tuple

from auto_editor.method import get_media_duration
from auto_editor.objects import VideoObj
from auto_editor.timeline import Timeline
from auto_editor.utils.func import to_timecode
from auto_editor.utils.log import Log


def time_frame(
    title: str, frames: float, fps: float, per: Optional[str] = None
) -> None:
    tc = to_timecode(frames / fps, "ass")

    tp = 9 if tc.startswith("-") else 10
    tcp = 12 if tc.startswith("-") else 11
    preci = 0 if int(frames) == frames else 2
    end = "" if per is None else f" {per:>7}"
    print(f" - {f'{title}:':<{tp}} {tc:<{tcp}} {f'({frames:.{preci}f})':<6}{end}")


def preview(timeline: Timeline, temp: str, log: Log) -> None:
    log.conwrite("")
    fps = timeline.fps

    # Calculate input videos length
    in_len = 0
    for inp in timeline.inputs:
        in_len += get_media_duration(inp.path, inp.get_fps(), temp, log)

    out_len = timeline.out_len()

    diff = out_len - in_len

    print("\nlength:")
    time_frame("input", in_len, fps, per="100.0%")
    time_frame("output", out_len, fps, per=f"{round((out_len / in_len) * 100, 2)}%")
    time_frame("diff", diff, fps, per=f"{round((diff / in_len) * 100, 2)}%")

    clip_lens = [clip.dur / clip.speed for clip in timeline.a[0]]

    # Calculate cuts
    oe: List[Tuple[int, int]] = []

    # TODO: Make offset_end_pairs work on overlapping clips.
    for clip in timeline.a[0]:
        oe.append((clip.offset, clip.offset + clip.dur))

    cut_lens = []
    i = 0
    while i < len(oe) - 1:
        if i == 0 and oe[i][0] != 0:
            cut_lens.append(oe[i][1])

        cut_lens.append(oe[i + 1][0] - oe[i][1])
        i += 1

    if len(oe) > 0 and oe[-1][1] < round(in_len * fps):
        cut_lens.append(in_len - oe[-1][1])

    log.debug(clip_lens)
    if len(clip_lens) == 0:
        clip_lens = [0]
    print(f"clips:\n - amount:    {len(clip_lens)}")
    time_frame("smallest", min(clip_lens), fps)
    time_frame("largest", max(clip_lens), fps)
    if len(clip_lens) > 1:
        time_frame("median", median(clip_lens), fps)
        time_frame("average", fmean(clip_lens), fps)

    log.debug(cut_lens)
    if len(cut_lens) == 0:
        cut_lens = [0]
    print(f"cuts:\n - amount:    {len(clip_lens)}")
    time_frame("smallest", min(cut_lens), fps)
    time_frame("largest", max(cut_lens), fps)
    if len(cut_lens) > 1:
        time_frame("median", median(cut_lens), fps)
        time_frame("average", fmean(cut_lens), fps)
    print("")



from typing import List, Tuple, Union, overload

import numpy as np
from numpy.typing import NDArray

from auto_editor.utils.log import Log
from auto_editor.utils.types import Chunks, frame_type

"""
To prevent duplicate code being pasted between scripts, common functions should be
put here. Every function should be pure with no side effects.
"""

# Turn long silent/loud array to formatted chunk list.
# Example: [1, 1, 1, 2, 2] => [(0, 3, 1), (3, 5, 2)]
def chunkify(arr: Union[np.ndarray, List[int]]) -> Chunks:
    arr_length = len(arr)

    chunks = []
    start = 0
    for j in range(1, arr_length):
        if arr[j] != arr[j - 1]:
            chunks.append((start, j, arr[j - 1]))
            start = j
    chunks.append((start, arr_length, arr[j]))
    return chunks


def to_timecode(secs: float, fmt: str) -> str:
    sign = ""
    if secs < 0:
        sign = "-"
        secs = -secs

    _m, _s = divmod(secs, 60)
    _h, _m = divmod(_m, 60)
    s, m, h = float(_s), int(_m), int(_h)

    if fmt == "webvtt":
        if h == 0:
            return f"{sign}{m:02d}:{s:06.3f}"
        return f"{sign}{h:02d}:{m:02d}:{s:06.3f}"
    if fmt == "mov_text":
        return f"{sign}{h:02d}:{m:02d}:" + f"{s:06.3f}".replace(".", ",", 1)
    if fmt == "standard":
        return f"{sign}{h:02d}:{m:02d}:{s:06.3f}"
    if fmt == "ass":
        return f"{sign}{h:d}:{m:02d}:{s:05.2f}"
    if fmt == "rass":
        return f"{sign}{h:d}:{m:02d}:{s:02.0f}"

    raise ValueError("to_timecode: Unreachable")


def remove_small(
    has_loud: NDArray[np.bool_], lim: int, replace: int, with_: int
) -> NDArray[np.bool_]:
    start_p = 0
    active = False
    for j, item in enumerate(has_loud):
        if item == replace:
            if not active:
                start_p = j
                active = True
            # Special case for end.
            if j == len(has_loud) - 1:
                if j - start_p < lim:
                    has_loud[start_p : j + 1] = with_
        else:
            if active:
                if j - start_p < lim:
                    has_loud[start_p:j] = with_
                active = False
    return has_loud


@overload
def set_range(
    arr: NDArray[np.float_],
    range_syntax: List[List[str]],
    fps: float,
    with_: int,
    log: Log,
) -> NDArray[np.float_]:
    pass


@overload
def set_range(
    arr: NDArray[np.bool_],
    range_syntax: List[List[str]],
    fps: float,
    with_: int,
    log: Log,
) -> NDArray[np.bool_]:
    pass


def set_range(arr, range_syntax, fps, with_, log):
    def replace_variables_to_values(val: str, fps: float, log: Log) -> int:
        if val == "start":
            return 0
        if val == "end":
            return len(arr)

        try:
            value = frame_type(val)
        except TypeError as e:
            log.error(f"{e}")
        if isinstance(value, int):
            return value
        return round(float(value) * fps)

    for _range in range_syntax:
        pair = []
        for val in _range:
            num = replace_variables_to_values(val, fps, log)
            if num < 0:
                num += len(arr)
            pair.append(num)
        arr[pair[0] : pair[1]] = with_
    return arr


def seconds_to_frames(value: Union[int, str], fps: float) -> int:
    if isinstance(value, str):
        return int(float(value) * fps)
    return value


def cook(has_loud: NDArray[np.bool_], min_clip: int, min_cut: int) -> NDArray[np.bool_]:
    has_loud = remove_small(has_loud, min_clip, replace=1, with_=0)
    has_loud = remove_small(has_loud, min_cut, replace=0, with_=1)
    return has_loud


def apply_margin(
    has_loud: NDArray[np.bool_], has_loud_length: int, start_m: int, end_m: int
) -> NDArray[np.bool_]:

    # Find start and end indexes.
    start_index = []
    end_index = []
    for j in range(1, has_loud_length):
        if has_loud[j] != has_loud[j - 1]:
            if has_loud[j]:
                start_index.append(j)
            else:
                end_index.append(j)

    # Apply margin
    if start_m > 0:
        for i in start_index:
            has_loud[max(i - start_m, 0) : i] = True
    if start_m < 0:
        for i in start_index:
            has_loud[i : min(i - start_m, has_loud_length)] = False

    if end_m > 0:
        for i in end_index:
            has_loud[i : min(i + end_m, has_loud_length)] = True
    if end_m < 0:
        for i in end_index:
            has_loud[max(i + end_m, 0) : i] = False

    return has_loud


def apply_mark_as(
    has_loud: NDArray[np.bool_], has_loud_length: int, fps: float, args, log: Log
) -> NDArray[np.bool_]:

    if args.mark_as_loud != []:
        has_loud = set_range(has_loud, args.mark_as_loud, fps, args.video_speed, log)

    if args.mark_as_silent != []:
        has_loud = set_range(has_loud, args.mark_as_silent, fps, args.silent_speed, log)
    return has_loud


def to_speed_list(
    has_loud: NDArray[np.bool_], video_speed: float, silent_speed: float
) -> NDArray[np.float_]:

    speed_list = has_loud.astype(float)

    # WARN: This breaks if speed is allowed to be 0
    speed_list[speed_list == 1] = video_speed
    speed_list[speed_list == 0] = silent_speed

    return speed_list


def merge(start_list: np.ndarray, end_list: np.ndarray) -> NDArray[np.bool_]:
    result = np.zeros((len(start_list)), dtype=np.bool_)

    for i, item in enumerate(start_list):
        if item == True:
            where = np.where(end_list[i:])[0]
            if len(where) > 0:
                result[i : where[0]] = True
    return result


def parse_dataclass(unsplit_arguments, dataclass, log):
    from dataclasses import fields

    # Positional Arguments
    #    --rectangle 0,end,10,20,20,30,#000, ...
    # Keyword Arguments
    #    --rectangle start=0,end=end,x1=10, ...

    ARG_SEP = ","
    KEYWORD_SEP = "="

    d_name = dataclass.__name__

    keys = [field.name for field in fields(dataclass)]
    kwargs = {}
    args = []

    allow_positional_args = True

    if unsplit_arguments == "":
        return dataclass()

    for i, arg in enumerate(unsplit_arguments.split(ARG_SEP)):
        if i + 1 > len(keys):
            log.error(f"{d_name} has too many arguments, starting with '{arg}'.")

        if KEYWORD_SEP in arg:
            allow_positional_args = False

            parameters = arg.split(KEYWORD_SEP)
            if len(parameters) > 2:
                log.error(f"{d_name} invalid syntax: '{arg}'.")
            key, val = parameters
            if key not in keys:
                log.error(f"{d_name} got an unexpected keyword '{key}'")

            kwargs[key] = val
        elif allow_positional_args:
            args.append(arg)
        else:
            log.error(f"{d_name} positional argument follows keyword argument.")

    try:
        dataclass_instance = dataclass(*args, **kwargs)
    except TypeError as err:
        err_list = [d_name] + str(err).split(" ")[1:]
        log.error(" ".join(err_list))

    return dataclass_instance


def get_stdout(cmd: List[str]) -> str:
    from subprocess import PIPE, STDOUT, Popen

    stdout, _ = Popen(cmd, stdout=PIPE, stderr=STDOUT).communicate()
    return stdout.decode("utf-8", "replace")


def aspect_ratio(width: int, height: int) -> Union[Tuple[int, int], Tuple[None, None]]:
    if height == 0:
        return None, None

    def gcd(a: int, b: int) -> int:
        while b:
            a, b = b, a % b
        return a

    c = gcd(width, height)
    return width // c, height // c


def human_readable_time(time_in_secs: float) -> str:
    units = "seconds"
    if time_in_secs >= 3600:
        time_in_secs = round(time_in_secs / 3600, 1)
        if time_in_secs % 1 == 0:
            time_in_secs = round(time_in_secs)
        units = "hours"
    if time_in_secs >= 60:
        time_in_secs = round(time_in_secs / 60, 1)
        if time_in_secs >= 10 or time_in_secs % 1 == 0:
            time_in_secs = round(time_in_secs)
        units = "minutes"
    return f"{time_in_secs} {units}"


def open_with_system_default(path: str, log: Log) -> None:
    import sys
    from subprocess import run

    if sys.platform == "win64" or sys.platform == "win32":
        from os import startfile

        try:
            startfile(path)
        except OSError:
            log.warning("Could not find application to open file.")
    else:
        try:  # should work on MacOS and some Linux distros
            run(["open", path])
        except Exception:
            try:  # should work on WSL2
                run(["cmd.exe", "/C", "start", path])
            except Exception:
                try:  # should work on most other Linux distros
                    run(["xdg-open", path])
                except Exception:
                    log.warning("Could not open output file.")


def append_filename(path: str, val: str) -> str:
    from os.path import splitext

    root, ext = splitext(path)
    return root + val + ext



import os.path
import subprocess
from dataclasses import dataclass
from typing import Dict, List, Tuple, Union

from auto_editor.ffwrapper import FFmpeg
from auto_editor.objects import EllipseObj, ImageObj, RectangleObj, TextObj, VideoObj
from auto_editor.output import get_vcodec, video_quality
from auto_editor.timeline import Timeline, Visual
from auto_editor.utils.encoder import encoders
from auto_editor.utils.log import Log
from auto_editor.utils.progressbar import ProgressBar


@dataclass
class VideoFrame:
    index: int
    source: int


# From: github.com/PyAV-Org/PyAV/blob/main/av/video/frame.pyx
allowed_pix_fmt = {
    "yuv420p",
    "yuvj420p",
    "rgb48be",
    "rgb48le",
    "rgb64be",
    "rgb64le",
    "rgb24",
    "bgr24",
    "argb",
    "rgba",
    "abgr",
    "bgra",
    "gray",
    "gray8",
    "gray16be",
    "gray16le",
    "rgb8",
    "bgr8",
    "pal8",
}


def apply_anchor(
    x: int, y: int, width: int, height: int, anchor: str
) -> Tuple[int, int]:
    if anchor == "ce":
        x = int((x * 2 - width) / 2)
        y = int((y * 2 - height) / 2)
    if anchor == "tr":
        x -= width
    if anchor == "bl":
        y -= height
    if anchor == "br":
        x -= width
        y -= height
    # Pillow uses 'tl' by default
    return x, y


def one_pos_two_pos(
    x: int, y: int, width: int, height: int, anchor: str
) -> Tuple[int, int, int, int]:
    """Convert: x, y, width, height -> x1, y1, x2, y2"""

    if anchor == "ce":
        x1 = x - int(width / 2)
        x2 = x + int(width / 2)
        y1 = y - int(height / 2)
        y2 = y + int(height / 2)

        return x1, y1, x2, y2

    if anchor in ("tr", "br"):
        x1 = x - width
        x2 = x
    else:
        x1 = x
        x2 = x + width

    if anchor in ("tl", "tr"):
        y1 = y
        y2 = y + height
    else:
        y1 = y
        y2 = y - height

    return x1, y1, x2, y2


def render_av(
    ffmpeg: FFmpeg,
    timeline: Timeline,
    args,
    progress: ProgressBar,
    rules,
    temp: str,
    log: Log,
) -> Tuple[str, bool]:
    try:
        import av

        av.logging.set_level(av.logging.PANIC)
    except ImportError:
        log.import_error("av")
    try:
        from PIL import Image, ImageChops, ImageDraw, ImageFont
    except ImportError:
        log.import_error("Pillow")

    font_cache: Dict[str, Union[ImageFont.FreeTypeFont, ImageFont.ImageFont]] = {}
    img_cache = {}
    for layer in timeline.v:
        for obj in layer:
            if isinstance(obj, TextObj) and obj.font not in font_cache:
                try:
                    if obj.font == "default":
                        font_cache[obj.font] = ImageFont.load_default()
                    else:
                        font_cache[obj.font] = ImageFont.truetype(obj.font, obj.size)
                except OSError:
                    log.error(f"Font '{obj.font}' not found.")

            if isinstance(obj, ImageObj) and obj.src not in img_cache:
                source = Image.open(obj.src)
                source = source.convert("RGBA")
                source = source.rotate(obj.rotate, expand=True)
                source = ImageChops.multiply(
                    source,
                    Image.new(
                        "RGBA", source.size, (255, 255, 255, int(obj.opacity * 255))
                    ),
                )
                img_cache[obj.src] = source

    inp = timeline.inp
    cn = av.open(inp.path, "r")
    pix_fmt = cn.streams.video[0].pix_fmt
    target_pix_fmt = pix_fmt if pix_fmt in allowed_pix_fmt else "yuv420p"
    my_codec = get_vcodec(args.video_codec, inp, rules)

    apply_video_later = True

    if my_codec in encoders:
        apply_video_later = encoders[my_codec]["pix_fmt"].isdisjoint(allowed_pix_fmt)

    if args.scale != 1:
        apply_video_later = False

    log.debug(f"apply video quality settings now: {not apply_video_later}")

    width = inp.videos[0].width
    height = inp.videos[0].height

    spedup = os.path.join(temp, "spedup0.mp4")

    cmd = [
        "-hide_banner",
        "-y",
        "-f",
        "rawvideo",
        "-c:v",
        "rawvideo",
        "-pix_fmt",
        target_pix_fmt,
        "-s",
        f"{width}*{height}",
        "-framerate",
        f"{timeline.fps}",
        "-i",
        "-",
        "-pix_fmt",
        target_pix_fmt,
    ]

    if apply_video_later:
        cmd.extend(["-c:v", "mpeg4", "-qscale:v", "1"])
    else:
        cmd = video_quality(cmd, args, inp, rules)

    cmd.append(spedup)

    process2 = ffmpeg.Popen(
        cmd, stdin=subprocess.PIPE, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )
    assert process2.stdin is not None

    stream = cn.streams.video[0]
    stream.thread_type = "AUTO"
    tou = int(stream.time_base.denominator / stream.average_rate)
    log.debug(f"Tou: {tou}")

    seek = 10
    seek_frame = None
    frames_saved = 0

    # Keyframes are usually spread out every 5 seconds or less.
    if args.no_seek:
        SEEK_COST = 4294967295
    else:
        SEEK_COST = int(stream.average_rate * 5)
    SEEK_RETRY = SEEK_COST // 2

    # New NLE method setup
    end = 0
    for layer in timeline.v:
        end = max(end, layer[-1].start + layer[-1].dur)

    d = cn.decode(stream)
    progress.start(end, "Creating new video")

    null_img = Image.new("RGB", (width, height), args.background)
    null_frame = av.VideoFrame.from_image(null_img).reformat(format=pix_fmt)

    frame_index = -1
    try:
        for index in range(end):
            # Add objects to obj_list
            obj_list: List[Union[VideoFrame, Visual]] = []
            for layer in timeline.v:
                for obj in layer:
                    if isinstance(obj, VideoObj):
                        if index >= obj.start and index < obj.start + round(
                            obj.speed * obj.dur
                        ):
                            obj_list.append(
                                VideoFrame(obj.offset + index - obj.start, obj.src)
                            )
                    elif index >= obj.start and index < obj.start + obj.dur:
                        obj_list.append(obj)

            # Render obj_list
            frame = null_frame
            for obj in obj_list:
                if isinstance(obj, VideoFrame):
                    if frame_index > obj.index:
                        log.debug("Seeking to beginning of file")
                        cn.seek(0)
                        frame = next(d)
                        frame_index = round(frame.time * timeline.fps)

                    while frame_index < obj.index:
                        # Check if skipping ahead is worth it.
                        if obj.index - frame_index > SEEK_COST and frame_index > seek:
                            seek = frame_index + SEEK_RETRY
                            seek_frame = frame_index
                            log.debug(f"Seeking to {frame_index * tou}")
                            cn.seek(frame_index * tou, stream=stream)

                        try:
                            frame = next(d)
                            frame_index = round(frame.time * timeline.fps)
                        except StopIteration:
                            log.warning(
                                "No source frame at current position. Using null frame"
                            )
                            frame = null_frame
                            break

                        if frame.key_frame:
                            log.debug(f"Keyframe {frame_index} {frame.pts}")
                        if seek_frame is not None:
                            log.debug(f"Skipped {frame_index - seek_frame} frames")
                            frames_saved += frame_index - seek_frame
                            seek_frame = None

                # Render visual objects
                else:
                    img = frame.to_image().convert("RGBA")
                    obj_img = Image.new("RGBA", img.size, (255, 255, 255, 0))
                    draw = ImageDraw.Draw(obj_img)

                    if isinstance(obj, TextObj):
                        text_w, text_h = draw.textsize(
                            obj.content, font=font_cache[obj.font]
                        )
                        pos = apply_anchor(obj.x, obj.y, text_w, text_h, "ce")
                        draw.text(
                            pos,
                            obj.content,
                            font=font_cache[obj.font],
                            fill=obj.fill,
                            align=obj.align,
                            stroke_width=obj.stroke,
                            stroke_fill=obj.strokecolor,
                        )

                    if isinstance(obj, RectangleObj):
                        draw.rectangle(
                            one_pos_two_pos(
                                obj.x, obj.y, obj.width, obj.height, obj.anchor
                            ),
                            fill=obj.fill,
                            width=obj.stroke,
                            outline=obj.strokecolor,
                        )

                    if isinstance(obj, EllipseObj):
                        draw.ellipse(
                            one_pos_two_pos(
                                obj.x, obj.y, obj.width, obj.height, obj.anchor
                            ),
                            fill=obj.fill,
                            width=obj.stroke,
                            outline=obj.strokecolor,
                        )

                    if isinstance(obj, ImageObj):
                        img_w, img_h = img_cache[obj.src].size
                        pos = apply_anchor(obj.x, obj.y, img_w, img_h, obj.anchor)
                        obj_img.paste(img_cache[obj.src], pos)

                    img = Image.alpha_composite(img, obj_img)
                    frame = frame.from_image(img).reformat(format=pix_fmt)

            process2.stdin.write(frame.to_ndarray().tobytes())

            if index % 3 == 0:
                progress.tick(index)

        progress.end()
        process2.stdin.close()
        process2.wait()
    except (OSError, BrokenPipeError):
        progress.end()
        ffmpeg.run_check_errors(cmd, log, True)
        log.error("FFmpeg Error!")

    log.debug(f"Total frames saved seeking: {frames_saved}")

    if args.scale != 1:
        sped_input = os.path.join(temp, "spedup0.mp4")
        spedup = os.path.join(temp, "scale0.mp4")

        cmd = [
            "-i",
            sped_input,
            "-vf",
            f"scale=iw*{args.scale}:ih*{args.scale}",
            spedup,
        ]

        check_errors = ffmpeg.pipe(cmd)
        if "Error" in check_errors or "failed" in check_errors:
            if "-allow_sw 1" in check_errors:
                cmd.insert(-1, "-allow_sw")
                cmd.insert(-1, "1")
            # Run again to show errors even if it might not work.
            ffmpeg.run(cmd)

    return spedup, apply_video_later



import os
import wave

from auto_editor.render.tsm.phasevocoder import phasevocoder
from auto_editor.timeline import Timeline
from auto_editor.utils.log import Log
from auto_editor.utils.progressbar import ProgressBar
from auto_editor.wavfile import read


def make_new_audio(
    t: int,
    temp: str,
    timeline: Timeline,
    progress: ProgressBar,
    log: Log,
) -> None:

    clips = timeline.a[t]
    if len(clips) == 0:
        log.error("Trying to create an empty file.")

    samples = []
    samplerate = 0
    for x in range(len(timeline.inputs)):
        samplerate, s = read(os.path.join(temp, f"{x}-{t}.wav"))
        samples.append(s)

    assert samplerate != 0

    progress.start(len(clips), "Creating new audio")
    fps = timeline.fps

    main_writer = wave.open(os.path.join(temp, f"new{t}.wav"), "wb")
    main_writer.setnchannels(2)
    main_writer.setframerate(samplerate)
    main_writer.setsampwidth(2)

    for c, clip in enumerate(clips):
        sample_start = int(clip.offset / fps * samplerate)
        sample_end = int((clip.offset + clip.dur) / fps * samplerate)

        samp_list = samples[clip.src]

        if sample_end > len(samp_list):
            sample_end = len(samp_list)

        if clip.speed == 1:
            main_writer.writeframes(samp_list[sample_start:sample_end])  # type: ignore
        else:
            output = phasevocoder(2, clip.speed, samp_list[sample_start:sample_end])
            if output.shape[0] != 0:
                main_writer.writeframes(output)  # type: ignore

        progress.tick(c)
    progress.end()



import os
import re
import sys
import tempfile

from auto_editor.ffwrapper import FFmpeg
from auto_editor.utils.log import Log
from auto_editor.vanparse import ArgumentParser


def grep_options(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument(
        "--no-filename", flag=True, help="Never print filenames with output lines."
    )
    parser.add_argument(
        "--max-count",
        "-m",
        type=int,
        help="Stop reading a file after NUM matching lines.",
    )
    parser.add_argument(
        "--count",
        "-c",
        flag=True,
        help="Suppress normal output; instead print count of matching lines for each file.",
    )
    parser.add_argument(
        "--ignore-case",
        "-i",
        flag=True,
        help="Ignore case distinctions for the PATTERN.",
    )
    parser.add_argument("--timecode", flag=True, help="Print the match's timecode.")
    parser.add_argument(
        "--time", flag=True, help="Print when the match happens. (Ignore ending)."
    )
    parser.add_argument("--ffmpeg-location", help="Point to your custom ffmpeg file.")
    parser.add_argument(
        "--my-ffmpeg",
        flag=True,
        help="Use the ffmpeg on your PATH instead of the one packaged.",
    )
    parser.add_required(
        "input", nargs="*", help="The path to a file you want inspected."
    )
    return parser


# stackoverflow.com/questions/9662346/python-code-to-remove-html-tags-from-a-string
def cleanhtml(raw_html: str) -> str:
    cleanr = re.compile("<.*?>")
    cleantext = re.sub(cleanr, "", raw_html)
    return cleantext


def grep_file(
    media_file: str, add_prefix: bool, ffmpeg: FFmpeg, args, log: Log, TEMP: str
) -> None:

    """
    We're using the WEBVTT subtitle format. It's better than srt
    because it doesn't emit line numbers and the time code is in
    (hh:mm:ss.sss) instead of (dd:hh:mm:ss,sss)
    """

    out_file = os.path.join(TEMP, "media.vtt")
    ffmpeg.run(["-i", media_file, out_file])

    count = 0

    flags = 0
    if args.ignore_case:
        flags = re.IGNORECASE

    prefix = ""
    if add_prefix:
        prefix = f"{os.path.splitext(os.path.basename(media_file))[0]}:"

    if args.max_count is None:
        args.max_count = float("inf")

    timecode = ""
    line_number = -1
    with open(out_file, "r") as file:
        while True:
            line = file.readline()
            line_number += 1
            if line_number == 0:
                continue

            if not line or count >= args.max_count:
                break

            if line.strip() == "":
                continue

            if re.match(r"\d*:\d\d.\d*\s-->\s\d*:\d\d.\d*", line):
                if args.time:
                    timecode = line.split("-->")[0].strip() + " "
                else:
                    timecode = line.strip() + "; "
                continue

            line = cleanhtml(line)
            match = re.search(args.input[0], line, flags)
            line = line.strip()

            if match:
                count += 1
                if not args.count:
                    if args.timecode or args.time:
                        print(prefix + timecode + line)
                    else:
                        print(prefix + line)

    if args.count:
        print(prefix + str(count))


def main(sys_args=sys.argv[1:]) -> None:
    parser = grep_options(ArgumentParser("grep"))
    args = parser.parse_args(sys_args)

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, debug=False)

    TEMP = tempfile.mkdtemp()
    log = Log(temp=TEMP)

    media_files = args.input[1:]
    add_prefix = (
        len(media_files) > 1 or os.path.isdir(media_files[0])
    ) and not args.no_filename

    for path in media_files:
        if os.path.isdir(path):
            for filename in [f for f in os.listdir(path) if not f.startswith(".")]:
                grep_file(
                    os.path.join(path, filename),
                    add_prefix,
                    ffmpeg,
                    args,
                    log,
                    TEMP,
                )
        elif os.path.isfile(path):
            grep_file(path, add_prefix, ffmpeg, args, log, TEMP)
        else:
            log.error(f"{path}: File does not exist.")

    log.cleanup()


if __name__ == "__main__":
    main()



"""
Export a FCPXML 9 file readable with Final Cut Pro 10.4.9 or later.

See docs here:
https://developer.apple.com/documentation/professional_video_applications/fcpxml_reference

"""

from pathlib import Path, PureWindowsPath
from platform import system
from typing import Union

from auto_editor.timeline import Timeline

from .utils import indent


def fraction(_a: Union[int, float], _fps: float) -> str:
    from fractions import Fraction

    if _a == 0:
        return "0s"

    if isinstance(_a, float):
        a = Fraction(_a)
    else:
        a = _a

    fps = Fraction(_fps)

    frac = Fraction(a, fps).limit_denominator()
    num = frac.numerator
    dem = frac.denominator

    if dem < 3000:
        factor = int(3000 / dem)

        if factor == 3000 / dem:
            num *= factor
            dem *= factor
        else:
            # Good enough but has some error that are impacted at speeds such as 150%.
            total = Fraction(0)
            while total < frac:
                total += Fraction(1, 30)
            num = total.numerator
            dem = total.denominator

    return f"{num}/{dem}s"


def fcp_xml(output: str, timeline: Timeline) -> None:
    inp = timeline.inp
    fps = timeline.fps
    chunks = timeline.chunks

    if chunks is None:
        raise ValueError("Timeline too complex")

    total_dur = chunks[-1][1]

    if system() == "Windows":
        pathurl = "file://localhost/" + PureWindowsPath(inp.abspath).as_posix()
    else:
        pathurl = Path(inp.abspath).as_uri()

    width, height = timeline.res
    frame_duration = fraction(1, fps)

    audio_file = len(inp.videos) == 0 and len(inp.audios) > 0
    group_name = "Auto-Editor {} Group".format("Audio" if audio_file else "Video")
    name = inp.basename

    with open(output, "w", encoding="utf-8") as outfile:
        outfile.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        outfile.write("<!DOCTYPE fcpxml>\n\n")
        outfile.write('<fcpxml version="1.9">\n')
        outfile.write("\t<resources>\n")
        outfile.write(
            f'\t\t<format id="r1" name="FFVideoFormat{height}p{fps}" '
            f'frameDuration="{frame_duration}" '
            f'width="{width}" height="{height}" '
            'colorSpace="1-1-1 (Rec. 709)"/>\n'
        )
        outfile.write(
            f'\t\t<asset id="r2" name="{name}" start="0s" hasVideo="1" format="r1" '
            'hasAudio="1" audioSources="1" audioChannels="2" '
            f'duration="{fraction(total_dur, fps)}">\n'
        )
        outfile.write(
            f'\t\t\t<media-rep kind="original-media" src="{pathurl}"></media-rep>\n'
        )
        outfile.write("\t\t</asset>\n")
        outfile.write("\t</resources>\n")
        outfile.write("\t<library>\n")
        outfile.write(f'\t\t<event name="{group_name}">\n')
        outfile.write(f'\t\t\t<project name="{name}">\n')
        outfile.write(
            indent(
                4,
                '<sequence format="r1" tcStart="0s" tcFormat="NDF" audioLayout="stereo" audioRate="48k">',
                "\t<spine>",
            )
        )

        last_dur = 0.0
        for clip in chunks:
            if clip[2] == 99999:
                continue

            clip_dur = (clip[1] - clip[0] + 1) / clip[2]
            dur = fraction(clip_dur, fps)

            close = "/" if clip[2] == 1 else ""

            if last_dur == 0:
                outfile.write(
                    indent(
                        6,
                        f'<asset-clip name="{name}" offset="0s" ref="r2" duration="{dur}" tcFormat="NDF"{close}>',
                    )
                )
            else:
                start = fraction(clip[0] / clip[2], fps)
                off = fraction(last_dur, fps)
                outfile.write(
                    indent(
                        6,
                        f'<asset-clip name="{name}" offset="{off}" ref="r2" '
                        + f'duration="{dur}" start="{start}" '
                        + f'tcFormat="NDF"{close}>',
                    )
                )

            if clip[2] != 1:
                # See the "Time Maps" section.
                # https://developer.apple.com/library/archive/documentation/FinalCutProX/Reference/FinalCutProXXMLFormat/StoryElements/StoryElements.html

                frac_total = fraction(total_dur, fps)
                speed_dur = fraction(total_dur / clip[2], fps)

                outfile.write(
                    indent(
                        6,
                        "\t<timeMap>",
                        '\t\t<timept time="0s" value="0s" interp="smooth2"/>',
                        f'\t\t<timept time="{speed_dur}" value="{frac_total}" interp="smooth2"/>',
                        "\t</timeMap>",
                        "</asset-clip>",
                    )
                )

            last_dur += clip_dur

        outfile.write("\t\t\t\t\t</spine>\n")
        outfile.write("\t\t\t\t</sequence>\n")
        outfile.write("\t\t\t</project>\n")
        outfile.write("\t\t</event>\n")
        outfile.write("\t</library>\n")
        outfile.write("</fcpxml>\n")



import os
from typing import List, Optional, Tuple

from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.timeline import Timeline, make_timeline
from auto_editor.utils.container import get_rules
from auto_editor.utils.log import Log
from auto_editor.utils.progressbar import ProgressBar
from auto_editor.utils.types import Chunk, Chunks


def set_output_name(path: str, inp_ext: str, export: str) -> str:
    root, ext = os.path.splitext(path)

    if export == "json":
        return f"{root}.json"
    if export == "final-cut-pro":
        return f"{root}.fcpxml"
    if export == "shotcut":
        return f"{root}.mlt"
    if export == "premiere":
        return f"{root}.xml"
    if export == "audio":
        return f"{root}_ALTERED.wav"
    if ext == "":
        return root + inp_ext

    return f"{root}_ALTERED{ext}"


def edit_media(
    path_list: List[str], ffmpeg: FFmpeg, args, temp: str, log: Log
) -> Optional[str]:

    progress = ProgressBar(args.progress)
    timeline = None

    path_ext = os.path.splitext(path_list[0])[1]
    if path_ext == ".json":
        from auto_editor.formats.json import read_json

        timeline = read_json(path_list[0], ffmpeg, log)
        inputs: List[FileInfo] = timeline.inputs
    else:
        inputs = []
        for path in path_list:
            inputs.append(FileInfo(path, ffmpeg, log))

    del path_list
    inp = inputs[0]

    if args.output_file is None:
        output = set_output_name(inp.path, inp.ext, args.export)
    else:
        output = args.output_file
        if os.path.splitext(output)[1] == "":
            output = set_output_name(output, inp.ext, args.export)

    output_container = os.path.splitext(output)[1].replace(".", "")

    # Check if export options make sense.
    rules = get_rules(output_container)
    codec_error = "'{}' codec is not supported in '{}' container."

    if not (args.sample_rate is None or args.sample_rate == "unset"):
        if rules.samplerate is not None and args.sample_rate not in rules.samplerate:
            log.error(
                f"'{output_container}' container only supports samplerates: {rules.samplerate}"
            )

    vcodec = args.video_codec
    if vcodec == "uncompressed":
        vcodec = "mpeg4"
    if vcodec == "copy":
        vcodec = inp.videos[0].codec

    if vcodec != "auto":
        if rules.vstrict:
            assert rules.vcodecs is not None
            if vcodec not in rules.vcodecs:
                log.error(codec_error.format(vcodec, output_container))

        if vcodec in rules.disallow_v:
            log.error(codec_error.format(vcodec, output_container))

    acodec = args.audio_codec
    if acodec == "copy":
        acodec = inp.audios[0].codec
        log.debug(f"Settings acodec to {acodec}")

    if acodec not in ("unset", "auto"):
        if rules.astrict:
            assert rules.acodecs is not None
            if acodec not in rules.acodecs:
                log.error(codec_error.format(acodec, output_container))

        if acodec in rules.disallow_a:
            log.error(codec_error.format(acodec, output_container))

    if args.keep_tracks_separate and rules.max_audios == 1:
        log.warning(
            f"'{output_container}' container doesn't support multiple audio tracks."
        )

    if not args.preview and not args.timeline:
        if os.path.isdir(output):
            log.error("Output path already has an existing directory!")

        if os.path.isfile(output) and inputs[0].path != output:
            log.debug(f"Removing already existing file: {output}")
            os.remove(output)

    # Extract subtitles in their native format.
    if len(inp.subtitles) > 0:
        cmd = ["-i", inp.path, "-hide_banner"]
        for s, sub in enumerate(inp.subtitles):
            cmd.extend(["-map", f"0:s:{s}"])
        for s, sub in enumerate(inp.subtitles):
            cmd.extend([os.path.join(temp, f"{s}s.{sub.ext}")])
        ffmpeg.run(cmd)

    log.conwrite("Extracting audio")

    cmd = []
    for i, inp in enumerate(inputs):
        cmd.extend(["-i", inp.path])
    cmd.append("-hide_banner")

    if args.sample_rate is None:
        samplerate = inputs[0].get_samplerate()
    else:
        samplerate = args.sample_rate

    for i, inp in enumerate(inputs):
        for s in range(len(inp.audios)):
            cmd.extend(
                [
                    "-map",
                    f"{i}:a:{s}",
                    "-ac",
                    "2",
                    "-ar",
                    f"{samplerate}",
                    "-rf64",
                    "always",
                    os.path.join(temp, f"{i}-{s}.wav"),
                ]
            )

    ffmpeg.run(cmd)

    if timeline is None:
        timeline = make_timeline(inputs, args, samplerate, progress, temp, log)

    if args.timeline:
        from auto_editor.formats.json import make_json_timeline

        make_json_timeline(args.api, 0, timeline, log)
        return None

    if args.preview:
        from auto_editor.preview import preview

        preview(timeline, temp, log)
        return None

    if args.export == "json":
        from auto_editor.formats.json import make_json_timeline

        make_json_timeline(args.api, output, timeline, log)
        return output

    if args.export == "premiere":
        from auto_editor.formats.premiere import premiere_xml

        premiere_xml(temp, output, timeline)
        return output

    if args.export == "final-cut-pro":
        from auto_editor.formats.final_cut_pro import fcp_xml

        fcp_xml(output, timeline)
        return output

    if args.export == "shotcut":
        from auto_editor.formats.shotcut import shotcut_xml

        shotcut_xml(output, timeline)
        return output

    def make_media(inp: FileInfo, timeline: Timeline, output: str) -> None:
        from auto_editor.output import mux_quality_media
        from auto_editor.render.video import render_av

        if rules.allow_subtitle:
            from auto_editor.render.subtitle import cut_subtitles

            cut_subtitles(ffmpeg, timeline, temp, log)

        if rules.allow_audio:
            from auto_editor.render.audio import make_new_audio

            for t in range(len(inp.audios)):
                make_new_audio(t, temp, timeline, progress, log)

        video_output = []

        if rules.allow_video:
            for v, vid in enumerate(inp.videos):
                if vid.codec not in ("png", "mjpeg", "webp") and v == 0:
                    out_path, apply_later = render_av(
                        ffmpeg, timeline, args, progress, rules, temp, log
                    )
                    video_output.append((v, True, out_path, apply_later))
                elif rules.allow_image:
                    out_path = os.path.join(temp, f"{v}.{vid.codec}")
                    # fmt: off
                    ffmpeg.run(["-i", inp.path, "-map", "0:v", "-map", "-0:V",
                        "-c", "copy", out_path])
                    # fmt: on
                    video_output.append((v, False, out_path, False))

        log.conwrite("Writing output file")

        mux_quality_media(
            ffmpeg, video_output, rules, output, output_container, args, inp, temp, log
        )

    if args.export == "clip-sequence":
        chunks = timeline.chunks
        if chunks is None:
            log.error("Timeline to complex to use clip-sequence export")

        total_frames = chunks[-1][1] - 1
        from auto_editor.timeline import clipify, make_av
        from auto_editor.utils.func import append_filename

        def pad_chunk(chunk: Chunk, total_frames: int) -> Chunks:
            start = [] if chunk[0] == 0 else [(0, chunk[0], 99999.0)]
            end = (
                [] if chunk[1] == total_frames else [(chunk[1], total_frames, 99999.0)]
            )
            return start + [chunk] + end

        clip_num = 0
        for chunk in chunks:
            if chunk[2] == 99999:
                continue

            _c = pad_chunk(chunk, total_frames)
            vspace, aspace = make_av(clipify(_c, 0), inp)
            my_timeline = Timeline(
                timeline.inputs,
                timeline.fps,
                timeline.samplerate,
                timeline.res,
                "#000",
                vspace,
                aspace,
                _c,
            )

            make_media(
                inp,
                my_timeline,
                append_filename(output, f"-{clip_num}"),
            )
            clip_num += 1
    else:
        make_media(inp, timeline, output)
    return output



import difflib
import re
import sys
import textwrap
from dataclasses import dataclass
from shutil import get_terminal_size
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union

import auto_editor
from auto_editor.utils.log import Log


@dataclass
class Required:
    names: Sequence[str]
    nargs: Union[int, str] = "*"
    type: type = str
    choices: Optional[Sequence[str]] = None
    help: str = ""
    _type: str = "required"


@dataclass
class Options:
    names: Sequence[str]
    nargs: Union[int, str] = 1
    type: type = str
    default: Optional[Union[int, str]] = None
    flag: bool = False
    choices: Optional[Sequence[str]] = None
    help: str = ""
    _type: str = "option"


@dataclass
class OptionText:
    text: str
    _type: str


def indent(text: str, prefix: str) -> str:
    def predicate(line: str) -> str:
        return line.strip()

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)

    return "".join(prefixed_lines())


def out(text: str) -> None:
    width = get_terminal_size().columns - 3

    indent_regex = re.compile(r"^(\s+)")
    wrapped_lines = []

    for line in text.split("\n"):
        exist_indent = re.search(indent_regex, line)
        pre_indent = exist_indent.groups()[0] if exist_indent else ""

        wrapped_lines.append(
            textwrap.fill(line, width=width, subsequent_indent=pre_indent)
        )

    print("\n".join(wrapped_lines))


def print_program_help(
    reqs: List[Required], args: List[Union[Options, OptionText]]
) -> None:
    text = ""
    for arg in args:
        if isinstance(arg, OptionText):
            text += f"\n  {arg.text}\n" if arg._type == "text" else "\n"
        else:
            text += "  " + ", ".join(arg.names) + f": {arg.help}\n"
    text += "\n"
    for req in reqs:
        text += "  " + ", ".join(req.names) + f": {req.help}\n"
    out(text)


def get_help_data() -> Dict[str, Dict[str, str]]:
    import json
    import os.path

    dirpath = os.path.dirname(os.path.realpath(__file__))

    with open(os.path.join(dirpath, "help.json"), "r") as fileobj:
        data = json.load(fileobj)

    assert isinstance(data, dict)
    return data


def print_option_help(program_name: str, option: Options) -> None:
    text = "  " + ", ".join(option.names) + f"\n    {option.help}\n\n"

    data = get_help_data()

    if option.names[0] in data[program_name]:
        text += indent(data[program_name][option.names[0]], "    ") + "\n\n"

    if option.flag:
        text += "    type: flag\n"
    else:
        text += f"    type: {option.type.__name__}\n"

        if option.nargs != 1:
            text += f"    nargs: {option.nargs}\n"

        if option.default is not None:
            text += f"    default: {option.default}\n"

        if option.choices is not None:
            text += "    choices: " + ", ".join(option.choices) + "\n"

    out(text)


def to_underscore(name: str) -> str:
    """Convert new style options to old style.  e.g. --hello-world -> --hello_world"""
    return name[:2] + name[2:].replace("-", "_")


def to_key(op: Union[Options, Required]) -> str:
    """Convert option name to arg key.  e.g. --hello-world -> hello_world"""
    return op.names[0][:2].replace("-", "") + op.names[0][2:].replace("-", "_")


def get_option(name: str, options: List[Options]) -> Optional[Options]:
    for option in options:
        if name in option.names or name in map(to_underscore, option.names):
            return option
    return None


class ArgumentParser:
    def __init__(self, program_name: str) -> None:
        self.program_name = program_name
        self.requireds: List[Required] = []
        self.options: List[Options] = []
        self.args: List[Union[Options, OptionText]] = []

    def add_argument(self, *args: str, **kwargs) -> None:
        x = Options(args, **kwargs)
        self.options.append(x)
        self.args.append(x)

    def add_required(self, *args: str, **kwargs) -> None:
        self.requireds.append(Required(args, **kwargs))

    def add_text(self, text: str) -> None:
        self.args.append(OptionText(text, "text"))

    def add_blank(self) -> None:
        self.args.append(OptionText("", "blank"))

    def parse_args(
        self,
        sys_args: List[str],
        macros: Optional[List[Tuple[Set[str], List[str]]]] = None,
    ):
        if len(sys_args) == 0:
            out(get_help_data()[self.program_name]["_"])
            sys.exit()

        if len(sys_args) == 1 and sys_args[0] in ("-v", "-V"):
            sys.stdout.write(f"{auto_editor.version} ({auto_editor.__version__})\n")
            sys.exit()

        if macros is not None:
            _macros = [(x[0].union(map(to_underscore, x[0])), x[1]) for x in macros]
            for old_options, new in _macros:
                for option in old_options:
                    if option in sys_args:
                        pos = sys_args.index(option)
                        sys_args[pos : pos + 1] = new

        return ParseOptions(
            sys_args, self.program_name, self.options, self.requireds, self.args
        )


class ParseOptions:
    @staticmethod
    def parse_value(option: Union[Options, Required], val: Optional[str]) -> Any:
        if val is None and option.nargs == 1:
            Log().error(f"{option.names[0]} needs argument.")

        try:
            value = option.type(val)
        except TypeError as e:
            Log().error(str(e))

        if option.choices is not None and value not in option.choices:
            my_choices = ", ".join(option.choices)

            Log().error(
                f"{value} is not a choice for {option.names[0]}\nchoices are:\n  {my_choices}"
            )

        return value

    def set_arg_list(self, key: str, my_list: List[str], list_type: type) -> None:
        try:
            setattr(self, key, list(map(list_type, my_list)))
        except (TypeError, ValueError) as e:
            Log().error(str(e))

    def __init__(
        self,
        sys_args: List[str],
        program_name: str,
        options: List[Options],
        requireds: List[Required],
        args: List[Union[Options, OptionText]],
    ) -> None:

        option_names: List[str] = []

        builtin_help = Options(
            ("--help", "-h"),
            flag=True,
            help="Show info about this program or option then exit.",
        )
        options.append(builtin_help)
        args.append(builtin_help)

        self.help = False

        # Set default attributes
        for op in options:
            for name in op.names:
                option_names.append(name)

            if op.flag:
                value: Any = False
            elif op.nargs != 1:
                value = []
            elif op.default is None:
                value = None
            else:
                value = op.type(op.default)

            setattr(self, to_key(op), value)

        # Figure out command line options changed by user.
        used_options: List[Options] = []

        req_list = []
        req_list_name = requireds[0].names[0]
        req_list_type = requireds[0].type
        setting_req_list = requireds[0].nargs != 1

        option_list = []
        op_list_name: Optional[str] = None
        op_list_type: type = str
        setting_op_list = False

        i = 0
        while i < len(sys_args):
            arg = sys_args[i]
            option = get_option(arg, options)

            if option is None:
                if setting_op_list:
                    option_list.append(arg)

                elif requireds and not arg.startswith("--"):

                    if requireds[0].nargs == 1:
                        setattr(
                            self, req_list_name, self.parse_value(requireds[0], arg)
                        )
                        requireds.pop()
                    else:
                        req_list.append(arg)
                else:
                    label = "option" if arg.startswith("--") else "short"

                    # 'Did you mean' message might appear that options need a comma.
                    if arg.replace(",", "") in option_names:
                        Log().error(f"Option '{arg}' has an unnecessary comma.")

                    close_matches = difflib.get_close_matches(arg, option_names)
                    if close_matches:
                        Log().error(
                            f"Unknown {label}: {arg}\n\n    Did you mean:\n        "
                            + ", ".join(close_matches)
                        )
                    Log().error(f"Unknown {label}: {arg}")
            else:
                if op_list_name is not None:
                    self.set_arg_list(op_list_name, option_list, op_list_type)

                if option in used_options:
                    Log().error(f"Cannot repeat option {option.names[0]} twice.")

                used_options.append(option)

                setting_op_list = False
                option_list = []
                op_list_name = None

                key = to_key(option)

                next_arg = None if i == len(sys_args) - 1 else sys_args[i + 1]
                if next_arg == "-h" or next_arg == "--help":
                    print_option_help(program_name, option)
                    sys.exit()

                if option.nargs != 1:
                    setting_op_list = True
                    op_list_name = key
                    op_list_type = option.type
                elif option.flag:
                    value = True
                else:
                    value = self.parse_value(option, next_arg)
                    i += 1
                setattr(self, key, value)

            i += 1

        if setting_op_list:
            assert op_list_name is not None
            self.set_arg_list(op_list_name, option_list, op_list_type)

        if setting_req_list:
            self.set_arg_list(req_list_name, req_list, req_list_type)

        if self.help:
            print_program_help(requireds, args)
            sys.exit()



import sys
from datetime import timedelta
from shutil import get_terminal_size, rmtree
from time import perf_counter, sleep
from typing import NoReturn, Optional


class Timer:
    __slots__ = ("start_time", "quiet")

    def __init__(self, quiet: bool = False) -> None:
        self.start_time = perf_counter()
        self.quiet = quiet

    def stop(self) -> None:
        if not self.quiet:
            second_len = round(perf_counter() - self.start_time, 2)
            minute_len = timedelta(seconds=round(second_len))

            sys.stdout.write(f"Finished. took {second_len} seconds ({minute_len})\n")


class Log:
    __slots__ = ("is_debug", "quiet", "temp")

    def __init__(
        self, show_debug: bool = False, quiet: bool = False, temp: Optional[str] = None
    ) -> None:
        self.is_debug = show_debug
        self.quiet = quiet
        self.temp = temp

    def debug(self, message: object) -> None:
        if self.is_debug:
            self.conwrite("")
            sys.stderr.write(f"Debug: {message}\n")

    def cleanup(self) -> None:
        if self.temp is None:
            return
        try:
            rmtree(self.temp)
            self.debug("Removed Temp Directory.")
        except FileNotFoundError:
            pass
        except PermissionError:
            sleep(0.1)
            try:
                rmtree(self.temp)
                self.debug("Removed Temp Directory.")
            except Exception:
                self.debug("Failed to delete temp dir.")

    def conwrite(self, message: str) -> None:
        if not self.quiet:
            buffer = " " * (get_terminal_size().columns - len(message) - 3)
            sys.stdout.write(f"  {message}{buffer}\r")

    def error(self, message: str) -> NoReturn:
        self.conwrite("")
        sys.stderr.write(f"Error! {message}\n")
        self.cleanup()
        from platform import system

        if system() == "Linux":
            sys.exit(1)
        else:
            try:
                sys.exit(1)
            except SystemExit:
                import os

                os._exit(1)

    def import_error(self, lib: str) -> NoReturn:
        self.error(f"Python module '{lib}' not installed. Run:  pip install {lib}")

    def warning(self, message: str) -> None:
        if not self.quiet:
            self.conwrite("")
            sys.stderr.write(f"Warning! {message}\n")

    def print(self, message: str) -> None:
        if not self.quiet:
            sys.stdout.write(f"{message}\n")
